import xarray as xr
import holoviews as hv
import panel as pn
from typing import List
import numpy as np


def convert_bool_dims_to_str(dataarray: xr.DataArray) -> xr.DataArray:
    """Given a dataarray that contains boolean coordinates, conver them to strings so that holoviews loads the data properly

    Args:
        dataarray (xr.DataArray): dataarray with boolean coordinates

    Returns:
        xr.DataArray: dataarray with boolean coordinates converted to strings
    """
    bool_coords = {}
    for c in dataarray.coords:
        if dataarray.coords[c].dtype == bool:
            bool_coords[c] = [str(vals) for vals in dataarray.coords[c].values]

    if len(bool_coords) > 0:
        return dataarray.assign_coords(bool_coords)
    return dataarray


def create_dataarray(power_on_coord_range: List[str | bool]) -> hv.Bars:
    """Create a 2D dataarray where voltage is measured when the power_on = True and power_on =False for two samples and plot with hv.Bars

    Args:
        power_on_coord_range (List[str|bool]): List values to use as a coordinate
    """
    da = xr.DataArray(
        name="voltage",
        data=[[1.5, 1.6], [0.01, -0.02]],
        coords=dict(power_on=(power_on_coord_range), sample_num=([1, 2])),
    )

    da = convert_bool_dims_to_str(da)

    hv_ds = hv.Dataset(da)
    print(f"datarray: {da}")
    print(f"dataframe: {da.to_dataframe()}")
    print(f"hv dataset: {hv_ds} ")
    return hv_ds.reduce(["sample_num"], np.mean).to(hv.Bars)


row = pn.Row()
# Create and plot a dataarray which uses strings as the index.  This works as expected
row.append(create_dataarray(power_on_coord_range=["True", "False"]))

# Create and plot the same data but use bools as the index.  This does not work because the values that have the index False do not appear in the holoviews dataset, even though they appear in the xarray dataset.  Bokeh warnings appear: BokehUserWarning: ColumnDataSource's columns must be of the same length. Current lengths: ('power_on', 1), ('voltage', 2).   These warnings occur because only data where power_on=True is included in the dataset but all the voltage data for both power_on=True and power_on=False is included
row.append(create_dataarray(power_on_coord_range=[True, False]))
row.show()



from collections import namedtuple
import xarray as xr
from sortedcontainers import SortedDict
import hashlib
import re
import math
from colorsys import hsv_to_rgb


def hmap_canonical_input(dic: dict) -> tuple:
    """From a dictionary of kwargs, return a hashable representation (tuple) that is always the same for the same inputs and retains the order of the input arguments.  e.g, {x=1,y=2} -> (1,2) and {y=2,x=1} -> (1,2).  This is used so that keywords arguments can be hashed and converted the the tuple keys that are used for holomaps

    Args:
        dic (dict): dictionary with keyword arguments and values in any order

    Returns:
        tuple: values of the dictionary always in the same order and hashable
    """

    function_input = SortedDict(dic)
    return tuple(function_input.values())


def make_namedtuple(class_name: str, **fields) -> namedtuple:
    """Convenience method for making a named tuple

    Args:
        class_name (str): name of the named tuple

    Returns:
        namedtuple: a named tuple with the fields as values
    """
    return namedtuple(class_name, fields)(*fields.values())


def get_nearest_coords(ds: xr.Dataset, **kwargs) -> dict:
    """Given an xarray dataset and kwargs of key value pairs of coordinate values, return a dictionary of the nearest coordinate name value pair that was found in the dataset

    Args:
        ds (xr.Dataset): dataset

    Returns:
        dict: nearest coordinate name value pair that matches the input coordinate name value pairs.
    """

    selection = ds.sel(method="nearest", **kwargs)
    cd = selection.coords.to_dataset().to_dict()["coords"]
    cd2 = {}
    for k, v in cd.items():
        cd2[k] = v["data"]
    return cd2


def hash_sha1(var: any) -> str:
    """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
    return hashlib.sha1(str(var).encode("ASCII")).hexdigest()


def capitalise_words(message: str):
    """Given a string of lowercase words, capitalise them

    Args:
        message (str): lower case string

    Returns:
        _type_: capitalised string
    """
    capitalized_message = " ".join([word.capitalize() for word in message.split(" ")])
    return capitalized_message


def un_camel(camel: str) -> str:
    """Given a snake_case string return a CamelCase string

    Args:
        camel (str): camelcase string

    Returns:
        str: uncamelcased string
    """

    return capitalise_words(re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", camel.replace("_", " ")))


def int_to_col(intVal, sat=0.5, val=0.95):
    golden_ratio_conjugate = (1 + math.sqrt(5)) / 2
    return hsv_to_rgb(intVal * golden_ratio_conjugate, sat, val)



import unittest
import bencher as bch
import random


class CachedParamExample(bch.CachedParams):
    var1 = bch.FloatSweep(default=0, bounds=[0, 10])
    var2 = bch.IntSweep(default=10, bounds=[0, 10])

    result = bch.ResultVar()

    def call(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)
        self.result = self.var1 + self.var2 + random.uniform(0, 1)
        return self.get_results_values_as_dict()

    def call_wrapped(self, **kwargs):
        return self.cache_wrap(self.call, **kwargs)


class TestCache(unittest.TestCase):
    def test_basic(self):
        cp = CachedParamExample()  # clears cache by default

        res1 = cp.call_wrapped(var1=1)
        res2 = cp.call_wrapped(var1=1)
        res3 = cp.call_wrapped(var1=2)
        res4 = cp.call_wrapped(var1=1, var2=10)  # calling with default value is the same
        res5 = cp.call_wrapped(var2=10, var1=1)  # calling with default value is the same

        # will only be equal if cache is used because of the randomness
        self.assertEqual(res1["result"], res2["result"])

        self.assertEqual(res1["result"], res4["result"])

        self.assertEqual(res4["result"], res5["result"])

        # will not be equal because called with different args
        self.assertNotEqual(res1["result"], res3["result"], f"{res1}")

        # create new class, make sure it has the same results
        cp2 = CachedParamExample(clear_cache=False)

        res1cp2 = cp2.call_wrapped(var1=1)

        self.assertEqual(res1["result"], res1cp2["result"])



# (Mostly) Generated by CodiumAI, sanity checked and fixed by a human
import unittest
from typing import Optional

import panel as pn

from bencher.plotting.plot_collection import PlotCollection, PlotProvider


class TestPlotProvider:
    def plot_1(self) -> Optional[pn.panel]:
        return [pn.pane.Markdown("Test plot 1", name="plot_1")]

    def plot_2(self) -> Optional[pn.panel]:
        return [pn.pane.Markdown("Test plot 2", name="plot_2")]


class TestPlotCollection(unittest.TestCase):
    # Tests that adding a plotter source successfully adds all the plotting methods of the class to the dictionary of available plotting functions
    def test_add_plotter_source_success(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        assert len(plot_coll.plotter_providers) == 2
        assert "plot_1" in plot_coll.plotter_providers
        assert "plot_2" in plot_coll.plotter_providers

    # Tests that adding a plot successfully adds the plot to the list of active plotting functions
    def test_add_plot_success(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        plot_coll.add("plot_1")
        assert len(plot_coll.plotters) == 1
        assert "plot_1" in plot_coll.plotters

    # Tests that removing a plot successfully removes the plot from the list of active plotting functions
    def test_remove_plot_success(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        plot_coll.add("plot_1")
        plot_coll.remove("plot_1")
        assert len(plot_coll.plotters) == 0

    # Tests that adding a plotter source with no plotting methods does not add any plotting methods to the dictionary of available plotting functions
    def test_add_plotter_source_no_methods(self):
        plot_coll = PlotCollection()

        class TestPlotProviderEmpty(PlotProvider):
            pass

        plot_coll.add_plotter_source(TestPlotProviderEmpty())
        assert len(plot_coll.plotter_providers) == 0

    # Tests that adding a non-existent plot does not add the plot to the list of active plotting functions
    def test_add_nonexistent_plot(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        with self.assertRaises(ValueError):
            plot_coll.add("plot_3")
        assert len(plot_coll.plotters) == 0

    # Tests that removing a non-existent plot does not remove any plot from the list of active plotting functions
    def test_remove_nonexistent_plot(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        plot_coll.add("plot_1")
        try:
            plot_coll.remove("plot_3")
        except KeyError:
            pass
        assert len(plot_coll.plotters) == 1

    def test_add_uninitialised_class(self):
        plot_coll = PlotCollection()
        with self.assertRaises(ValueError):
            plot_coll.add_plotter_source(TestPlotProvider)

    def test_add_not_a_list(self):
        plot_coll = PlotCollection()

        plot_coll.add_plotter_source(TestPlotProvider())
        with self.assertRaises(ValueError):
            # try adding a variable that is not a list
            plot_coll.add_list("plot_1")

    def test_add_list_empty(self):
        plot_coll = PlotCollection()
        plot_coll.add_plotter_source(TestPlotProvider())
        plot_coll.add_list([])
        assert len(plot_coll.plotters) == 0



from typing import Optional
import panel as pn
import hvplot.xarray  # noqa pylint: disable=unused-import
import holoviews as hv  # noqa pylint: disable=unused-import

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
from bencher.plotting.plot_types import PlotTypes  # noqa pylint: disable=unused-import


class HvInteractive:
    float_0_cat_at_least1_vec_1_res_1 = PlotFilter(
        float_range=VarRange(0, 0),
        cat_range=VarRange(1, None),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    # shared plot filter for catplots
    scatter_filter = PlotFilter(
        float_range=VarRange(0, 0),
        cat_range=VarRange(1, None),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    lineplot_filter = PlotFilter(
        float_range=VarRange(1, None),
        cat_range=VarRange(0, 0),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    lineplot_multi_filter = PlotFilter(
        float_range=VarRange(2, None),
        cat_range=VarRange(0, 0),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(-1, -1),
    )

    def bar_hv(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if self.scatter_filter.matches(pl_in.plt_cnt_cfg):
            pt = pl_in.bench_cfg.to_bar()
            return pn.Column(pt, name=PlotTypes.bar_hv)

        return None

    def scatter_hv(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if self.scatter_filter.matches(pl_in.plt_cnt_cfg):
            pt = pl_in.bench_cfg.to_scatter_jitter()
            # pt *= pl_in.bench_cfg.get_hv_dataset(False).to(hv.Scatter).opts(color="k", jitter=0.5)
            return pn.Column(pt, name=PlotTypes.scatter_hv)

        return None

        # return None
        # if self.scatter_filter.matches(pl_in.plt_cnt_cfg):
        #     x = pl_in.plt_cnt_cfg.cat_vars[0]
        #     y = pl_in.rv
        #     ds = pl_in.bench_cfg.ds[pl_in.rv.name]
        #     title = f"{x.name} vs {y.name}"

        #     if pl_in.plt_cnt_cfg.cat_cnt > 1:
        #         c = pl_in.plt_cnt_cfg.cat_vars[1].name
        #         ds = ds.set_index(c)

        #     pt = hv.Overlay()

        #     pt *= ds.mean("repeat").hvplot(x=x.name, y=y.name, kind="bar")
        #     for r in ds.coords["repeat"]:
        #         pt *= ds.sel(repeat=r).hvplot(
        #             x=x.name, y=y.name, kind="scatter", rot=45, color="black"
        #         )
        #     # pt = pl_in.bench_cfg.to_points()

        #     return pn.Column(pt.opts(height=600, title=title), name=PlotTypes.scatter_hv)

        # return None

    def lineplot_hv(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if False & self.lineplot_filter.matches(pl_in.plt_cnt_cfg):
            # print(pl_in.bench_cfg.get_hv_dataset())
            # print(pl_in.bench_cfg.get_dataframe(False))
            # return pn.Column(pl_in.bench_cfg.get_hv_dataset().to(hv.Table))
            return pn.Column(pl_in.bench_cfg.to_curve(), name=PlotTypes.lineplot_hv)
        return None

    def lineplot_hv_overlay(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if False & self.lineplot_multi_filter.matches(pl_in.plt_cnt_cfg):
            return pn.Column(
                pl_in.bench_cfg.to_curve().overlay(pl_in.bench_cfg.input_vars[-1].name),
                name=PlotTypes.lineplot_hv_overlay,
            )
        return None

    def lineplot_hv_layout(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if False & self.lineplot_multi_filter.matches(pl_in.plt_cnt_cfg):
            return pn.Column(
                pl_in.bench_cfg.to_curve().layout(),
                name=PlotTypes.lineplot_hv_layout,
            )
        return None



from hypothesis import given, settings
from hypothesis import strategies as st

from bencher.bench_cfg import PltCntCfg
from bencher.example.benchmark_data import ExampleBenchCfgOut
from bencher.plotting.plot_collection import PlotInput
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting.plots.tables import Tables

from .test_plots_common import TestPlotsCommon


class TestTables(TestPlotsCommon):
    @settings(deadline=10000)
    @given(
        st.sampled_from(
            [
                PlotTypes.dataframe_flat,
                PlotTypes.dataframe_mean,
                PlotTypes.dataframe_multi_index,
                PlotTypes.xarray,
            ]
        )
    )
    def test_plot_name(self, plot_name) -> None:
        bench_cfg = self.create_bench_cfg(plot_name)

        plt_cnt_cfg = PltCntCfg(float_cnt=0, cat_cnt=1)
        pl_in = PlotInput(bench_cfg, ExampleBenchCfgOut.param.out_cos, plt_cnt_cfg)
        cp = Tables()
        plot_fn = getattr(cp, plot_name)
        result = plot_fn(pl_in)

        self.basic_plot_asserts(result, plot_name)



from __future__ import annotations
from typing import List, Any, Tuple
from copy import deepcopy

import numpy as np
import param
from param import Parameterized
import holoviews as hv
import panel as pn
from bencher.utils import hash_sha1

# slots that are shared across all Sweep classes
# param does not work with multiple inheritance so define here
shared_slots = ["units", "samples", "samples_debug"]


def hash_extra_vars(parameter: Parameterized) -> int:
    """hash extra meta vars in the parameter

    Args:
        parameter (Parameterized): a parameter

    Returns:
        int: hash
    """
    return hash_sha1((parameter.units, parameter.samples, parameter.samples_debug))


def describe_variable(
    v: Parameterized, debug: bool, include_samples: bool, value=None
) -> List[str]:
    """Generate a string description of a variable

    Args:
        v (param.Parameterized): parameter to describe
        debug (bool): Generate a reduced number of samples from the variable
        include_samples (bool): Include a description of the samples

    Returns:
        str: String description of the variable
    """
    indent = "    "
    sampling_str = []
    sampling_str.append(f"{v.name}:")
    if include_samples:
        # sampling_str.append(f"{indent}{v.sampling_str(debug)}")
        sampling_str.append(f"{indent}number of samples: {len(v.values(debug))}")
        sampling_str.append(f"{indent}sample values: {v.values(debug)}")

    if value is not None:
        sampling_str.append(f"{indent}value: {value}")
    if v.units != "ul":
        sampling_str.append(f"{indent}units: [{v.units}]")
    if v.doc is not None:
        sampling_str.append(f"{indent}docs: {v.doc}")
    for i in range(len(sampling_str)):
        sampling_str[i] = f"{indent}{sampling_str[i]}"
    return sampling_str


class SweepBase(param.Parameter):
    # def __init__(self, **params):
    # super().__init__(**params)
    # self.units = ""
    # slots = ["units", "samples", "samples_debug"]
    # __slots__ = shared_slots

    def values(self, debug: bool) -> List[Any]:
        """All sweep classes must implement this method. This generates sample values from based on the parameters bounds and sample number.

        Args:
            debug (bool): Return a reduced set of samples to enable fast debugging of a data generation and plotting pipeline. Ideally when debug is true, 2 samples will be returned

        Returns:
            List[Any]: A list of samples from the variable
        """
        raise NotImplementedError

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_extra_vars(self)

    def sampling_str(self, debug=False) -> str:
        """Generate a string representation of the of the sampling procedure

        Args:
            debug (bool): If true then self.samples_debug is used
        """

        samples = self.values(debug)
        object_str = ",".join([str(i) for i in samples])
        return f"Taking {len(samples)} samples from {self.name} with values: [{object_str}]"

    def as_slider(self, debug=False) -> pn.widgets.slider.DiscreteSlider:
        """given a sweep variable (self), return the range of values as a panel slider

        Args:
            debug (bool, optional): pass to the sweepvar to produce a full set of varaibles, or when debug=True, a reduces number of sweep vars. Defaults to False.

        Returns:
            pn.widgets.slider.DiscreteSlider: A panel slider with the values() of the sweep variable
        """
        return pn.widgets.slider.DiscreteSlider(name=self.name, options=list(self.values(debug)))

    def as_dim(self, compute_values=False, debug=False) -> hv.Dimension:
        """Takes a sweep variable and turns it into a holoview dimension

        Returns:
            hv.Dimension:
        """
        name_tuple = (self.name, self.name)

        params = {}
        if hasattr(self, "bounds"):
            if compute_values:
                params["values"] = self.values(debug)
                # params["range"] = tuple(self.bounds)
            else:
                params["range"] = tuple(self.bounds)
                params["default"] = self.default

        else:
            params["values"] = self.values(debug)

        return hv.Dimension(name_tuple, unit=self.units, **params)  # pylint: disable=no-member

    def indices_to_samples(self, desires_num_samples, sample_values):
        indices = [
            int(i) for i in np.linspace(0, len(sample_values) - 1, desires_num_samples, dtype=int)
        ]

        if len(indices) > len(sample_values):
            return sample_values

        return [sample_values[i] for i in indices]

    def with_samples(self, samples: int) -> SweepBase:
        output = deepcopy(self)
        # TODO set up class properly. Slightly complicated due to slots
        output.samples = samples  # pylint: disable = attribute-defined-outside-init
        return output

    def with_sample_values(self, sample_values: int) -> SweepBase:
        output = deepcopy(self)
        # TODO set up class properly. Slightly complicated due to slots
        try:
            output.sample_values = sample_values  # pylint: disable = attribute-defined-outside-init
        except AttributeError:
            output.objects = sample_values  # pylint: disable = attribute-defined-outside-init
        return output

    def with_const(self, const_value: Any) -> Tuple[SweepBase, Any]:
        """Create a new instance of SweepBase with a constant value.

        Args:
            const_value (Any): The constant value to be associated with the new instance.

        Returns:
            Tuple[SweepBase, Any]: A tuple containing the new instance of SweepBase and the constant value.
        """
        return (deepcopy(self), const_value)

    def with_level(self, level: int = 1, max_level: int = 12) -> SweepBase:
        assert level >= 1
        # TODO work out if the order can be returned in level order always
        samples = [0, 1, 2, 3, 5, 9, 17, 33, 65, 129, 257, 513, 1025, 2049]
        return self.with_samples(samples[min(max_level, level)])



"""This file has some examples for how to perform basic benchmarking parameter sweeps"""
# pylint: disable=duplicate-code


from bencher.bencher import Bench, BenchRunCfg

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


def example_time_event(run_cfg: BenchRunCfg) -> Bench:
    """This example shows how to manually set time events as a string so that progress can be monitored over time"""

    bencher = Bench("benchmarking_example_categorical1D", bench_function, ExampleBenchCfgIn)

    ExampleBenchCfgIn.param.offset.bounds = [0, 100]

    # manually override the default value based on the time event string so that the graphs are not all just straight lines
    ExampleBenchCfgIn.param.offset.default = int(str(hash(run_cfg.time_event))[-1])

    # here we sample the input variable theta and plot the value of output1. The (noisy) function is sampled 20 times so you can see the distribution
    bencher.plot_sweep(
        title="Example 1D Categorical",
        input_vars=[ExampleBenchCfgIn.param.postprocess_fn],
        result_vars=[ExampleBenchCfgOut.param.out_cos],
        description=example_time_event.__doc__,
        run_cfg=run_cfg,
    )
    return bencher


def run_example_time_event(ex_run_cfg):
    ex_run_cfg.repeats = 1
    ex_run_cfg.print_pandas = True
    ex_run_cfg.over_time = True

    ex_run_cfg.clear_cache = True
    ex_run_cfg.clear_history = True

    ex_run_cfg.time_event = "-first_event"
    example_time_event(ex_run_cfg)

    ex_run_cfg.clear_cache = False
    ex_run_cfg.clear_history = False
    ex_run_cfg.time_event = "_second_event"
    example_time_event(ex_run_cfg)

    ex_run_cfg.time_event = (
        "*third_event has a very very long label to demonstrate automatic text wrapping"
    )
    return example_time_event(ex_run_cfg)


if __name__ == "__main__":
    run_example_time_event.show()



from __future__ import annotations

from dataclasses import dataclass

from bencher.bench_cfg import BenchCfg, PltCntCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep


class VarRange:
    def __init__(self, lower_bound: int = 0, upper_bound: int = -1) -> None:
        """A VarRange represents the bounded and unbounded ranges of integers.  This class is used to define filters for various variable types.  For example by defining cat_var = VarRange(0,0), calling matches(0) will return true, but any other integer will not match.  You can also have unbounded ranges for example VarRange(2,None) will match to 2,3,4... up to infinity. for By default the lower and upper bounds are set to -1 so so that no matter what value is passsed to matches() will return false.  Matches only takes 0 and positive integers.

        Args:
            lower_bound (int, optional): The smallest acceptable value to matches(). Passing None will result in a lower bound of 0 (as matches only accepts positive integers). Defaults to 0.
            upper_bound (int, optional): The largest acceptable value to matches().  Passing None will result in no upper bound. Defaults to -1.
        """
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def matches(self, val: int) -> bool:
        """Checks that a value is within the variable range.  lower_bound and upper_bound are inclusive (lower_bound<=val<=upper_bound )

        Args:
            val (int): A positive integer representing a number of items

        Returns:
            bool: True if the items is within the range, False otherwise.

        Raises:
            ValueError: If val < 0
        """
        if val < 0:
            raise ValueError("val must be >= 0")
        if self.lower_bound is not None:
            lower_match = val >= self.lower_bound
        else:
            lower_match = True

        if self.upper_bound is not None:
            upper_match = val <= self.upper_bound
        else:
            upper_match = True

        return lower_match and upper_match

    def __repr__(self) -> str:
        return f"VarRange(lower_bound={self.lower_bound}, upper_bound={self.upper_bound})"


class PlotFilter:
    def __init__(
        self,
        float_range: VarRange = VarRange(),
        cat_range: VarRange = VarRange(),
        vector_len: VarRange = VarRange(1, 1),
        result_vars: VarRange = VarRange(1, 1),
    ) -> None:
        """A class for representing the types of results a plot is able to represent.

        Args:
            float_range (VarRange, optional): The range of float varibles the plot supports. Defaults to VarRange().
            cat_range (VarRange, optional): The range of categorical varibles the plot supports. Defaults to VarRange().
            vector_len (VarRange, optional): The range of vector lengths the plot supports. Defaults to VarRange(1, 1).
            result_vars (VarRange, optional): The range of number of result varibles the plot supports. Defaults to VarRange(1, 1).
        """

        self.float_range = float_range
        self.cat_range = cat_range
        self.vector_len = vector_len
        self.result_vars = result_vars

    def matches(self, plt_cng_cfg: PltCntCfg) -> bool:
        """Checks if the result data signature matches the type of data the plot is able to display.

        Args:
            plt_cng_cfg (PltCntCfg): The plot count configuration to check.

        Returns:
            bool: True if the configuration matches the filter, False otherwise.
        """

        return (
            self.float_range.matches(plt_cng_cfg.float_cnt)
            and self.cat_range.matches(plt_cng_cfg.cat_cnt)
            and self.vector_len.matches(plt_cng_cfg.vector_len)
            and self.result_vars.matches(plt_cng_cfg.result_vars)
        )


@dataclass
class PlotInput:
    """A dataclass that contains all the information needed to plot
    Args:
        bench_cfg (BenchCfg): The benchmark configuration used to generate the result data
        rv (ParametrizedSweep): The result variable to plot
        plt_cnt_cfg (PltCntCfg): The number and types of variable to plot
    """

    bench_cfg: BenchCfg
    rv: ParametrizedSweep
    plt_cnt_cfg: PltCntCfg


class PlotProvider:
    """A base class for code that displays or plots data. Each class that inherits provides plotting methods and a filter that specifies what the plot is capable of displaying"""

    # commonly used filters that are shared across classes
    float_1_cat_any_vec_1_res_1_ = PlotFilter(
        float_range=VarRange(0, 0),
        cat_range=VarRange(0, None),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )



# from bencher import Pa#
import param


from holoviews import opts
import holoviews as hv

hv.extension("bokeh", "plotly")

width_heigh = {"width": 600, "height": 600, "tools": ["hover"]}

opts.defaults(
    opts.Curve(**width_heigh),
    opts.Points(**width_heigh),
    opts.Bars(**width_heigh),
    opts.Scatter(**width_heigh),
    opts.HeatMap(cmap="plasma", **width_heigh, colorbar=True),
    # opts.Surface(**width_heigh),
    opts.GridSpace(plot_size=400),
)


class PlotBase:
    def title(self, x: param.Parameter, y: param.Parameter, z: param.Parameter = None) -> str:
        if z is None:
            return f"{x.name} vs {y.name}"
        return f"{z.name} vs ({x.name} vs {y.name})"



# import holoviews as hv
# import numpy as np
# import param
# from holoviews import opts, streams
# from holoviews.plotting.links import DataLink

# import panel as pn

# hv.extension("bokeh")

# datasets = {
#     "data1": np.random.randint(1, 10, size=(4, 2)),
#     "data2": np.random.randint(1, 10, size=(4, 2)),
# }

# other_data = {i: np.random.randint(1, 10, size=(100, 100)) for i in range(4)}


# class Explorer(param.Parameterized):
#     dataset = param.ObjectSelector(default=list(datasets.keys())[0], objects=list(datasets.keys()))

#     lines = param.Boolean(default=False)

#     # streams
#     stream = param.ClassSelector(
#         default=streams.PointDraw(drag=False, add=False), class_=(streams.PointDraw), precedence=-1
#     )

#     selection = param.ClassSelector(
#         default=streams.Selection1D(), class_=(streams.Selection1D), precedence=-1
#     )

#     @param.depends("dataset")
#     def load_file(self):
#         name = self.dataset
#         datum = datasets[name]
#         scatter = hv.Scatter(datum).opts(size=8)

#         # update the PointDraw/Selection1D sources/data
#         self.stream.source = scatter
#         self.selection.source = scatter
#         self.stream.update(data=scatter.columns())  # reset PointDraw data
#         self.selection.update(index=[])  # reset selection index

#         table = hv.Table(scatter).opts(index_position=False, editable=True, selectable=True)
#         DataLink(scatter, table)

#         return (scatter + table).opts(
#             opts.Scatter(tools=["tap", "hover", "box_select"]), opts.Table(editable=True)
#         )

#     @param.depends("dataset", "selection.index", "lines")
#     def view(self):
#         """update 3rd plot whenever dataset, selection.index, or lines is changed"""
#         if self.selection.index == []:
#             return None
#         else:
#             # modify with your "other" data
#             if self.lines is True:
#                 return hv.Path(other_data[self.selection.index[0]]).opts(shared_axes=False)

#         return hv.operation.datashader.regrid(
#             hv.Image(other_data[self.selection.index[0]]),
#             upsample=True,
#             interpolation="bilinear",
#         ).opts(shared_axes=False)

#     @param.depends("stream.data", watch=True)
#     def update_data(self):
#         """update dataset whenever a point is deleted using PointDraw"""
#         datasets[self.dataset] = np.vstack((self.stream.data["x"], self.stream.data["y"])).T


# explorer = Explorer()
# pn.Row(pn.Column(explorer.param, explorer.view), explorer.load_file).show()



from typing import List, Tuple, Any

import param
from param import Parameterized
import holoviews as hv
from bencher.utils import make_namedtuple, hash_sha1
from bencher.variables.results import ResultVar, ResultVec
from functools import partial


class ParametrizedSweep(Parameterized):
    """Parent class for all Sweep types that need a custom hash"""

    @staticmethod
    def param_hash(
        param_type: Parameterized, hash_value: bool = True, hash_meta: bool = False
    ) -> int:
        """A custom hash function for parametrised types with options for hashing the value of the type and hashing metadata

        Args:
            param_type (Parameterized): A parameter
            hash_value (bool, optional): use the value as part of the hash. Defaults to True.
            hash_meta (bool, optional): use metadata as part of the hash. Defaults to False.

        Returns:
            int: a hash
        """

        curhash = 0
        if hash_value:
            for k, v in param_type.param.values().items():
                if k != "name":
                    curhash = hash_sha1((curhash, hash_sha1(v)))

        if hash_meta:
            for k, v in param_type.param.params().items():
                if k != "name":
                    print(f"key:{k}, hash:{hash_sha1(k)}")
                    print(f"value:{v}, hash:{hash_sha1(v)}")
                    curhash = hash_sha1((curhash, hash_sha1(k), hash_sha1(v)))
        return curhash

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return ParametrizedSweep.param_hash(self, True, False)

    def update_params_from_kwargs(self, **kwargs) -> None:
        """Given a dictionary of kwargs, set the parameters of the passed class 'self' to the values in the dictionary."""
        used_params = {}
        for key in self.param.params().keys():
            if key in kwargs:
                if key != "name":
                    used_params[key] = kwargs[key]

        self.param.update(**used_params)

    @classmethod
    def get_input_and_results(cls, include_name: bool = False) -> Tuple[dict, dict]:
        """Get dictionaries of input parameters and result parameters

        Args:
            cls: A parametrised class
            include_name (bool): Include the name parameter that all parametrised classes have. Default False

        Returns:
            Tuple[dict, dict]: a tuple containing the inputs and result parameters as dictionaries
        """
        inputs = {}
        results = {}
        for k, v in cls.param.params().items():
            if isinstance(v, (ResultVar, ResultVec)):
                results[k] = v
            else:
                inputs[k] = v

        if not include_name:
            inputs.pop("name")
        return make_namedtuple("inputresult", inputs=inputs, results=results)

    def get_inputs_as_dict(self) -> dict:
        """Get the key:value pairs for all the input variables"""
        inp = self.get_input_and_results().inputs
        vals = self.param.values()
        return {i: vals[i] for i, v in inp.items()}

    def get_results_values_as_dict(self, holomap=None) -> dict:
        """Get a dictionary of result variables with the name and the current value"""
        values = self.param.values()
        output = {key: values[key] for key in self.get_input_and_results().results}
        if holomap is not None:
            output |= {"hmap": holomap}
        return output

    @classmethod
    def get_inputs_only(cls) -> List[param.Parameter]:
        """Return a list of input parameters

        Returns:
            List[param.Parameter]: A list of input parameters
        """
        return list(cls.get_input_and_results().inputs.values())

    @staticmethod
    def filter_fn(item, p_name):
        return item.name != p_name

    @classmethod
    def get_input_defaults(cls, override_defaults=None) -> List[Tuple[param.Parameter, Any]]:
        inp = cls.get_inputs_only()
        if override_defaults is None:
            override_defaults = []
        assert isinstance(override_defaults, list)
        for p in override_defaults:
            inp = filter(partial(ParametrizedSweep.filter_fn, p_name=p[0].name), inp)
        return override_defaults + [(i, i.default) for i in inp]

    @classmethod
    def get_results_only(cls) -> List[param.Parameter]:
        """Return a list of input parameters

        Returns:
            List[param.Parameter]: A list of result parameters
        """
        return list(cls.get_input_and_results().results.values())

    def get_inputs_as_dims(
        self, compute_values=False, remove_dims: str | List[str] = None
    ) -> List[hv.Dimension]:
        inputs = self.get_inputs_only()

        if remove_dims is not None:
            if isinstance(remove_dims, str):
                remove_dims = [remove_dims]
            filtered_inputs = [i for i in inputs if i.name not in remove_dims]
            inputs = filtered_inputs

        return [iv.as_dim(compute_values) for iv in inputs]

    def to_dynamic_map(
        self,
        callback=None,
        name=None,
        remove_dims: str | List[str] = None,
    ) -> hv.DynamicMap:
        if callback is None:
            callback = self.call

        def callback_wrapper(**kwargs):
            return callback(**kwargs)["hmap"]

        return hv.DynamicMap(
            callback=callback_wrapper,
            kdims=self.get_inputs_as_dims(compute_values=False, remove_dims=remove_dims),
            name=name,
        ).opts(shared_axes=False, framewise=True, width=1000, height=1000)

    def to_holomap(self, callback, remove_dims: str | List[str] = None) -> hv.DynamicMap:
        return hv.HoloMap(
            hv.DynamicMap(
                callback=callback,
                kdims=self.get_inputs_as_dims(compute_values=True, remove_dims=remove_dims),
            )
        )
        # return hv.HoloMap(self.to_dynamic_map(callback=callback, remove_dims=remove_dims))
        # return hv.DynamicMap(
        #     kdims=self.get_inputs_as_dims(compute_values=True, remove_dims=remove_dims)
        # )

    def call(self):
        pass

    def plot_hmap(self, **kwargs):
        return self.call(**kwargs)["hmap"]



import unittest

from bencher.example.benchmark_data import AllSweepVars, PostprocessFn
import bencher as bch
from hypothesis import given, strategies as st


class TestSweepBase(unittest.TestCase):
    def test_with_samples(self) -> None:
        """Check that using with_samples does not have side effects"""

        sweep_samples_before = AllSweepVars.param.var_float.values()
        custom_samples = AllSweepVars.param.var_float.with_samples(5).values()
        sweep_samples_after = AllSweepVars.param.var_float.values()

        self.assertEqual(str(sweep_samples_before), str(sweep_samples_after))
        self.assertNotEqual(str(sweep_samples_before), str(custom_samples))

    def test_with_const(self) -> None:
        """Check that setting a const returns the right const"""

        res = AllSweepVars.param.var_float.with_const(5)

        self.assertEqual(res[1], 5)

    def test_setting_const(self) -> None:
        """Check that setting a const returns the right const"""

        explorer = AllSweepVars()
        bench = bch.Bench("tst_cnst", explorer.__call__)

        # AllSweepVars.param.var_float.with_const(5)

        consts = explorer.get_input_defaults()
        consts_len = len(consts)

        res = bench.plot_sweep(
            "tst",
            input_vars=[AllSweepVars.param.var_float.with_samples(3)],
            const_vars=consts,
        )

        consts_after = [i[0] for i in res.const_vars]

        self.assertEqual(consts_len, len(consts))
        self.assertEqual(consts_len - 1, len(consts_after))

        self.assertTrue(AllSweepVars.param.var_float not in consts)
        self.assertTrue(explorer.param.var_float not in consts)

    def test_override_const(self) -> None:
        """Check that setting a const returns the right const"""

        explorer = AllSweepVars()
        bch.Bench("tst_cnst", explorer.call)

        consts = explorer.get_input_defaults()
        const_override = explorer.get_input_defaults([AllSweepVars.param.var_float.with_const(2)])

        # const_override_class = AllSweepVars.get_input_defaults(
        #     [AllSweepVars.param.var_float.with_const(2)]
        # )

        # const_override_instance = explorer.get_input_defaults(
        #     [explorer.param.var_float.with_const(2)]
        # )

        print(consts)
        print("overridden")
        print(const_override)

        self.assertEqual(consts[0][1], 5)
        self.assertEqual(const_override[0][1], 2)
        # self.assertEqual(const_override_class[0][1], 2)
        # self.assertEqual(const_override_instance[0][1], 2)

        self.assertNotEqual(consts, const_override)
        self.assertEqual(len(consts), len(const_override))

    def test_override_defaults(self):
        exp = AllSweepVars()

        class_defaults = AllSweepVars.get_input_defaults(
            [AllSweepVars.param.var_float.with_const(3)]
        )
        self.assertEqual(class_defaults[0][1], 3)

        # check that the defaults have not been modified
        default_defaults = AllSweepVars.get_input_defaults()
        self.assertEqual(default_defaults[0][1], 5)

        instance_defaults = exp.get_input_defaults([exp.param.var_float.with_const(2)])
        self.assertEqual(instance_defaults[0][1], 2)

    def test_with_sample_values(self):
        vals = AllSweepVars.param.var_float.with_sample_values([0, 1]).values(False)
        self.assertEqual(vals[0], 0)
        self.assertEqual(vals[1], 1)

        defaults = AllSweepVars.param.var_float.values(False)
        self.assertEqual(defaults[9], 10)

        vals = AllSweepVars.param.var_enum.with_sample_values([PostprocessFn.negate]).values(False)
        self.assertEqual(len(vals), 1)
        self.assertEqual(vals[0], PostprocessFn.negate)

    def test_bool_as_dim(self):
        res = AllSweepVars.param.var_bool.as_dim(True)

        self.assertSequenceEqual(res.values, [True, False])

        res = AllSweepVars.param.var_bool.as_dim(False)
        self.assertSequenceEqual(res.values, [True, False])

    def test_float_as_dim(self):
        res = AllSweepVars.param.var_float.as_dim(True)

        self.assertListEqual(res.values, list(AllSweepVars.param.var_float.values()))

        res = AllSweepVars.param.var_float.as_dim(False)
        self.assertSequenceEqual(res.range, (0, 10))

    def sweep_up_to(self, var, var_type, level=7):
        res_old = var.with_level(1)
        for i in range(2, level):
            res = var.with_level(i)
            new_vals = res.values()
            print(res_old.values(), new_vals)
            for i in res_old.values():
                self.assertTrue(isinstance(i, var_type))
                self.assertTrue(i in new_vals)
                for n in new_vals:
                    print("\t", i == n)
            res_old = res

    @given(st.floats(min_value=0.1, allow_nan=False, allow_infinity=False))
    def test_levels_float(self, upper) -> None:
        var_float = bch.FloatSweep(bounds=(0, upper))
        self.sweep_up_to(var_float, float)

    # @given(st.integers(min_value=0), st.integers(min_value=1,max_value=10))
    # def test_levels_int(self, start, var_range):
    #     var_int = bch.IntSweep(default=start, bounds=(start, start + var_range))
    #     self.sweep_up_to(var_int, int, level=5)


if __name__ == "__main__":
    # TestSweepBase().test_override_defaults()

    TestSweepBase().test_levels_int(0, 10)



import bencher as bch
from bencher.utils import int_to_col

import math
import holoviews as hv
from typing import Any, List
import panel as pn
from holoviews import opts


class LevelsExample(bch.ParametrizedSweep):
    xval = bch.FloatSweep(bounds=[0, 3.14])
    yval = bch.FloatSweep(bounds=[0, 3.14])
    level = bch.IntSweep(default=1, bounds=[1, 6])

    output = bch.ResultVar(units="v")

    def __call__(self, **kwargs: Any) -> Any:
        self.update_params_from_kwargs(**kwargs)
        self.output = math.sin(self.xval) + math.cos(self.yval)
        pt = hv.Points((self.xval, self.yval)).opts(
            marker="o", size=110 - self.level * 20, color=int_to_col(self.level - 1)
        )

        return self.get_results_values_as_dict(pt)


class RunWithLevel(bch.ParametrizedSweep):
    level = bch.IntSweep(default=1, bounds=[1, 8])
    dimensions = bch.IntSweep(default=1, bounds=[1, 2])

    level_samples = bch.ResultVar()

    def __call__(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)

        # bench_level = bch.Bench(
        #     "lvl", LevelsExample(), run_cfg=bch.BenchRunCfg(level=self.level, auto_plot=False)
        # )

        # res = bench.plot_sweep(
        #     f"Level:{self.level}",
        #     input_vars=dims,
        #     const_vars=LevelsExample.get_input_defaults(
        #         [LevelsExample.param.level.with_const(self.level)]
        #     ),
        #     result_vars=[LevelsExample.param.output],
        #     run_cfg=bch.BenchRunCfg(level=self.level, auto_plot=False),
        # )

        self.level_samples = int(
            pow(
                len(bch.FloatSweep(bounds=[0, 1]).with_level(self.level).values(False)),
                self.dimensions,
            )
        )
        return self.get_results_values_as_dict()

        # return self.get_results_values_as_dict(res1.to_holomap().overlay())


# def to_bench(class_instance) -> bch.Bench:
#     bench = bch.Bench(f"bench_{class_instance.name}", class_instance)
#     bench.plot_sweep(f"bench_{class_instance.name}")
#     return bench


def run_with_dim(bench: bch.Bench, dims: List[bch.SweepBase]):
    results = []
    for level in range(1, 6):
        print(level)
        res = bench.plot_sweep(
            f"Level:{level}",
            input_vars=dims,
            const_vars=LevelsExample.get_input_defaults(
                [LevelsExample.param.level.with_const(level)]
            ),
            result_vars=[LevelsExample.param.output],
            run_cfg=bch.BenchRunCfg(level=level, auto_plot=False),
        )

        results.append(res)
    return results


def run_levels_1D(bench: bch.Bench) -> bch.Bench:
    results = run_with_dim(bench, [LevelsExample.param.xval])
    bench.append_markdown("# Using Levels to define sample density", "Levels")

    bench1 = bch.Bench("lol", RunWithLevel(), run_cfg=bch.BenchRunCfg(auto_plot=False))
    res1 = bench1.plot_sweep("Levels", input_vars=[RunWithLevel.param.level])

    bench.append_markdown(
        "Sample levels let you perform parameter sweeps without having to decide how many samples to take.  If you perform a sweep at level 1, then all the points are reused when sampling at level 2.  The higher levels reuse the points from lower levels to avoid having to recompute potentially expensive samples. The other advantage is that it enables a workflow where you can quickly see the results of the sweep at a low resolution to sense check the code, and then run it at a high level to get the fidelity you want.  When calling a sweep at a high level, you can publish the intermediate lower level results as the computiation continues so that you can track the progress of the computation and end the sweep early when you have sufficient resolution",
        width=600,
    )
    row = pn.Row()
    row.append(res1.to_table())
    row.append(res1.to_curve().opts(shared_axes=False))
    bench.append(row)

    bench.append_markdown(
        "Level 1 returns a single point at the lower bound of the parameter. Level 2 uses the uppper and lower bounds of the parameter. All subsequent levels are created by adding a sample between each previously calculated sample to ensure that all previous values can be reused while retaining an equal sample spacing.  The following plots show the sample points as circles and the corresponding plot of a sin function sampled at that level.",
        width=600,
    )

    combined_pts = hv.Overlay()
    combined_curve = hv.Overlay()
    for it, r in enumerate(results):
        lvl = it + 1
        row = pn.Row()
        pts = r.to_holomap().overlay().opts(title=f"Sample Points for level: {lvl}", height=300)
        crv = r.to_curve().opts(shared_axes=False, height=300) * r.to_hv_dataset().to(
            hv.Scatter
        ).opts(title=f"Function Values for level: {lvl}", size=5, height=300, shared_axes=False)

        combined_pts *= pts
        combined_curve *= crv
        row.append(pts)
        row.append(crv)
        bench.append_markdown(f"## {r.title}")
        bench.append(row)

    bench.append_markdown(
        "This plot overlays the previous plots into a single image. It shows how each level overlaps the previous level"
    )

    bench.append(pn.Row(combined_pts, combined_curve))
    return bench


def run_levels_2D(bench: bch.Bench) -> bch.Bench:
    results = run_with_dim(bench, [LevelsExample.param.xval, LevelsExample.param.yval])
    bench.append_markdown("# Using Levels to define 2D sample density", "Levels 2D")

    bench1 = bch.Bench("lol", RunWithLevel(), run_cfg=bch.BenchRunCfg(auto_plot=False))
    res1 = bench1.plot_sweep(
        "Levels",
        input_vars=[RunWithLevel.param.level],
        const_vars=[RunWithLevel.param.dimensions.with_const(2)],
    )
    row = pn.Row()
    row.append(res1.to_table())
    row.append(res1.to_curve().opts(shared_axes=False))
    bench.append(row)

    for it, r in enumerate(results):
        lvl = it + 1
        row = pn.Row()
        bench.append_markdown(f"## {r.title}")
        row.append(
            r.to_holomap()
            .overlay()
            .opts(title=f"Sample Points for level: {lvl}", shared_axes=False)
        )
        row.append(
            r.to_heatmap().opts(title=f"Function Value Heatmap for level: {lvl}", shared_axes=False)
        )
        bench.append(row)

    bench.append_markdown(
        "This plot overlays the previous plots into a single image. It shows how each level overlaps the previous level"
    )
    overlay = hv.Overlay()
    for lvl, r in enumerate(results):
        overlay *= (
            r.to_holomap()
            .overlay()
            .opts(width=1000, height=1000, show_legend=False, shared_axes=False)
        )

    bench.append(overlay)
    return bench


def run_levels() -> bch.Bench:
    hv.extension("bokeh")
    opts.defaults(
        opts.Curve(show_legend=False),
        opts.Points(show_legend=False),
    )

    bench = bch.Bench("Levels", LevelsExample())
    bench = run_levels_1D(bench)
    bench = run_levels_2D(bench)

    return bench


if __name__ == "__main__":
    run_levels().show()



import bencher as bch
import unittest
from bencher.utils import get_nearest_coords

import xarray as xr


class ExampleClass(bch.ParametrizedSweep):
    iv1 = bch.FloatSweep()
    rv1 = bch.ResultVar()


class TestBencherUtils(unittest.TestCase):
    def test_get_inputs(self) -> None:
        ex_instance = ExampleClass()
        inputs = ex_instance.get_inputs_only()

        print(inputs)

        self.assertEqual(len(inputs), 1)
        self.assertEqual(inputs[0].name, "iv1")

    def test_get_results(self) -> None:
        ex_instance = ExampleClass()
        results = ex_instance.get_results_only()

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].name, "rv1")

    def test_get_inputs_and_results(self) -> None:
        ex_instance = ExampleClass()
        inputs, results = ex_instance.get_input_and_results()

        self.assertEqual(len(inputs), 1)
        self.assertEqual(len(results), 1)
        self.assertEqual(inputs["iv1"].name, "iv1")
        self.assertEqual(results["rv1"].name, "rv1")

    def test_get_results_values_as_dict(self) -> None:
        ex_instance = ExampleClass()

        ex_instance.rv1 = 3

        res = ex_instance.get_results_values_as_dict()

        self.assertEqual(res["rv1"], 3)

        # Tests that a named tuple with fields of different data types is created successfully

    def test_different_datatypes_namedtuple(self):
        result = bch.make_namedtuple("Test", field1=1, field2="value2", field3=True)
        self.assertEqual(result.field1, 1)
        self.assertEqual(result.field2, "value2")
        self.assertEqual(result.field3, True)

        # Tests that the function returns an empty tuple when an empty dictionary is passed as input

    def test_edge_case_empty_dictionary(self) -> None:
        input_dict = {}
        expected_output = ()
        self.assertEqual(bch.hmap_canonical_input(input_dict), expected_output)

    def test_dictionary_order(self) -> None:
        dic1 = {"x": 1, "y": 2}
        dic2 = {"y": 2, "x": 1}

        self.assertEqual(
            bch.hmap_canonical_input(dic1),
            bch.hmap_canonical_input(dic2),
        )

    # Tests that the function returns the nearest coordinate name value pair for a dataset containing multiple coordinates
    def test_multiple_coordinates(self):
        ds = xr.Dataset({"x": [1, 2, 3], "y": [4, 5, 6]})
        result = get_nearest_coords(ds, x=2.5, y=5.5)
        self.assertEqual(result, {"x": 3, "y": 6})



from textwrap import wrap

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import panel as pn
import plotly.express as px
import seaborn as sns

from bencher.bench_cfg import BenchCfg, PltCfgBase
from bencher.variables.parametrised_sweep import ParametrizedSweep

from bencher.variables.results import ResultVar, ResultVec


# hv.extension("plotly", "bokeh")
# hv.extension("plotly")


def wrap_long_time_labels(bench_cfg: BenchCfg) -> BenchCfg:
    """Takes a benchCfg and wraps any index labels that are too long to be plotted easily

    Args:
        bench_cfg (BenchCfg):

    Returns:
        BenchCfg: updated config with wrapped labels
    """
    if bench_cfg.over_time:
        if bench_cfg.ds.coords["over_time"].dtype == np.datetime64:
            # plotly catastrophically fails to plot anything with the default long string representation of time, so convert to a shorter time representation
            bench_cfg.ds.coords["over_time"] = [
                pd.to_datetime(t).strftime("%d-%m-%y %H-%M-%S")
                for t in bench_cfg.ds.coords.coords["over_time"].values
            ]
            # wrap very long time event labels because otherwise the graphs are unreadable
        if bench_cfg.time_event is not None:
            bench_cfg.ds.coords["over_time"] = [
                "\n".join(wrap(t, 20)) for t in bench_cfg.ds.coords["over_time"].values
            ]
    return bench_cfg


def plot_sns(bench_cfg: BenchCfg, rv: ParametrizedSweep, sns_cfg: PltCfgBase) -> pn.pane:
    """Plot with seaborn

    Args:
        bench_cfg (BenchCfg): bench config
        rv (ParametrizedSweep): the result variable to plot
        sns_cfg (PltCfgBase): the plot configuration

    Returns:
        pn.pane: A seaborn plot as a panel pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    plt.rcParams.update({"figure.max_open_warning": 0})

    # if type(rv) == ResultVec:
    # return plot_scatter_sns(bench_cfg, rv)
    # else:

    if type(rv) == ResultVec:
        if rv.size == 2:
            plt.figure(figsize=(4, 4))
            fg = plot_scatter2D_sns(bench_cfg, rv)
        else:
            return pn.pane.Markdown("Scatter plots of >3D result vectors not supported yet")
    elif type(rv) == ResultVar:
        df = bench_cfg.ds[rv.name].to_dataframe().reset_index()

        try:
            fg = sns_cfg.plot_callback(data=df, **sns_cfg.as_sns_args())
        except Exception as e:
            return pn.pane.Markdown(
                f"Was not able to plot becuase of exception:{e} \n this is likely due to too many NAN values"
            )

        # TODO try to set this during the initial plot rather than after
        for ax in fg.axes.flatten():
            for tick in ax.get_xticklabels():
                tick.set_rotation(45)

        fg.set_xlabels(label=sns_cfg.xlabel, clear_inner=True)
        fg.set_ylabels(label=sns_cfg.ylabel, clear_inner=True)

    fg.fig.suptitle(sns_cfg.title)
    plt.tight_layout()

    return pn.panel(plt.gcf())


def plot_scatter2D_sns(bench_cfg: BenchCfg, rv: ParametrizedSweep) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D scatter plot from seaborn

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot

    Returns:
        pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)
    ds = bench_cfg.ds.drop_vars("repeat")

    df = ds.to_dataframe().reset_index()

    names = rv.index_names()

    if bench_cfg.input_vars:
        h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.input_vars[0].name)
    elif bench_cfg.over_time:
        h = sns.jointplot(df, x=names[0], y=names[1], hue=bench_cfg.iv_time[0].name)

    else:
        h = sns.jointplot(df, x=names[0], y=names[1])

    h.set_axis_labels(f"{names[0]} [{rv.units}]", f"{names[1]} [{rv.units}]")
    return h


def plot_scatter2D_hv(bench_cfg: BenchCfg, rv: ParametrizedSweep) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D scatter plot

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot

    Returns:
        pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)
    bench_cfg.ds.drop_vars("repeat")

    df = bench_cfg.get_dataframe()

    names = rv.index_names()

    return px.scatter(df, x=names[0], y=names[1], marginal_x="histogram", marginal_y="histogram")



from typing import Optional

import panel as pn
import plotly.express as px
import xarray as xr

import hvplot.xarray  # noqa pylint: disable=unused-import


from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange
from bencher.plotting.plot_types import PlotTypes
from bencher.variables.parametrised_sweep import ParametrizedSweep


class Heatmap:
    # shared plot filter for 2d plots
    plot_filter = PlotFilter(
        float_range=VarRange(2, 2),
        cat_range=VarRange(0, 0),
        vector_len=VarRange(1, 1),
        result_vars=VarRange(1, 1),
    )

    def imshow_wrapper(
        self,
        df: xr.DataArray,
        x: ParametrizedSweep,
        y: ParametrizedSweep,
        z: ParametrizedSweep,
        name: str,
    ) -> pn.panel:
        """given a dataframe and result variables, use imshow to plot a heatmap

        Args:
            df (xr.DataArray): data to plot
            x (ParametrizedSweep): x variable
            y (ParametrizedSweep): y variable
            z (ParametrizedSweep): z variable
            name (str): name of the plot

        Returns:
            pn.panel: A heatmap
        """
        title = f"{z.name} vs ({x.name} vs {y.name})"
        xlabel = f"{x.name} [{x.units}]"
        ylabel = f"{y.name} [{y.units}]"
        color_label = f"{z.name} [{z.units}]"

        # TODO set up a holoview version
        # return pn.panel(df.hvplot.heatmap(x=x.name, y=y.name, C=z.name, colorbar=True))
        return pn.panel(
            px.imshow(
                df,
                title=title,
                labels={"x": xlabel, "y": ylabel, "color": color_label},
            ),
            name=name,
        )

    def heatmap_2D(self, pl_in: PlotInput) -> Optional[pn.panel]:
        """use the imshow plotting method to display 2D data

        Args:
            pl_in (PlotInput): The data to plot

        Returns:
            pn.panel: A panel with a image representation of the data
        """
        if len(pl_in.bench_cfg.input_vars) == 2:
            da = pl_in.bench_cfg.ds[pl_in.rv.name]
            mean = da.mean("repeat")

            x = pl_in.bench_cfg.input_vars[0]
            y = pl_in.bench_cfg.input_vars[1]
            z = pl_in.rv

            title = f"{z.name} vs ({x.name} vs {y.name})"
            color_label = f"{z.name} [{z.units}]"

            return pn.panel(
                mean.hvplot.heatmap(
                    x=x.name,
                    y=y.name,
                    C=z.name,
                    title=title,
                    label=pl_in.rv.name,
                    colorbar=True,
                    clabel=color_label,
                    cmap="viridis",
                ),
                name=PlotTypes.heatmap_2D,
            )
        return None

    def heatmap_1D(self, pl_in: PlotInput) -> Optional[pn.panel]:
        """use the imshow plotting method to display a heatmap

        Args:
            pl_in (PlotInput): The data to plot

        Returns:
            Optional[pn.panel]: A panel with a image representation of the data
        """
        if len(pl_in.bench_cfg.input_vars) == 1:
            da = pl_in.bench_cfg.ds[pl_in.rv.name]
            mean = da.mean("repeat", keepdims=True, keep_attrs=True)

            return [
                self.imshow_wrapper(
                    da,
                    pl_in.bench_cfg.iv_repeat,
                    pl_in.bench_cfg.input_vars[0],
                    pl_in.rv,
                    PlotTypes.heatmap_2D,
                ),
                self.imshow_wrapper(
                    mean,
                    pl_in.bench_cfg.iv_repeat,
                    pl_in.bench_cfg.input_vars[0],
                    pl_in.rv,
                    PlotTypes.heatmap_1D,
                ),
            ]
        return None



import bencher as bch

from example_investment import Investment
from example_mortgage import Mortgage


class PersonalFinances(bch.ParametrizedSweep):
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=4)

    net_income = bch.FloatSweep(default=3000, bounds=(0, 5000))
    # savings = bch.FloatSweep(default=1000, bounds=(0, 3000))
    # mortgage_payment = bch.FloatSweep(default=1000, bounds=(500, 2000))

    # monthly_contribution = bch.FloatSweep(default=0, bounds=[0, 1000], units="$", samples=4)
    # investment_interest = bch.FloatSweep(default=6.0, bounds=(0.01, 10), units="%", samples=4)
    # tax_rate = bch.IntSweep(default=0, bounds=(0, 40), samples=2)
    # period = bch.IntSweep(default=12, bounds=[1, 12], units="month")
    # starting_value = bch.FloatSweep(default=1000, bounds=(0, 2000), samples=3)

    # discretionary_income = bch.FloatSweep(default=500, bounds=(0, 1000), units="$")
    # investment_to_mortgage_ratio = bch.FloatSweep(default=0.3, bounds=(0, 1), units="ratio")
    # investments_value = bch.ResultVar(default=0, units="$")

    # results
    net_wealth = bch.ResultVar(default=0, units="$")
    investments_value = bch.ResultVar(default=0, units="$")

    def call(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)

        inv = Investment(**kwargs)
        mort = Mortgage(**kwargs)

        pt = inv.plot_hmap()
        pt *= mort.plot_hmap()

        return self.get_results_values_as_dict(pt.opts(legend_position="right"))

        # # mort_inv = self.net_income - self.discretionary_income
        # # self.monthly_contribution = (
        # #     self.net_income
        # #     - self.discretionary_income
        # #     # - self.installment_payment
        # # )
        # # self.mortgage_payment
        # # self.monthly_contribution = (
        # #     self.post_tax_income * self.investment_to_mortgage_ratio
        # # )

        # for i in range(self.period):
        #     self.investments_value += self.monthly_contribution
        #     new_interest = self.investments_value * (self.investment_interest / 100.0)
        #     self.investments_value += new_interest * (1 + self.tax_rate)
        #     self.investments_value *= 1 - (self.inflation / 100.0)
        #     values.append((i, self.investments_value))

        # # self.calc_loan(**kwargs)

        # self.net_wealth + self.investments_value + self.installment_principal
        # print(values)
        # return self.get_results_values_as_dict(
        #     hv.Curve(values)
        #     # hv.Curve(values)
        # )


if __name__ == "__main__":
    pf = PersonalFinances()
    bench = bch.Bench("PersonalFinances", pf.call, plot_lib=None)

    # res = bench.plot_sweep(
    #     "Mortgage",
    #     [pf.param.interest],
    #     [
    #         pf.param.installment_balance,
    #         pf.param.installment_principal,
    #     ],
    # )

    # bench.append(res.to_curve())

    # bench.append(res.to(hv.Table).opts(width=1000))

    # bench.append(res.to_holomap())

    # bench.append(res.to_grid())

    bench.append_tab(pf.to_dynamic_map(name="Mortgage Calculator"))

    bench.show()



# pylint: disable=duplicate-code

import numpy as np

import bencher as bch


class VolumeSample(bch.ParametrizedSweep):
    """A class to represent a 3D point in space."""

    x = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="x coordinate of the sample volume", samples=4
    )
    y = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="y coordinate of the sample volume", samples=5
    )
    z = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="z coordinate of the sample volume", samples=6
    )


class VolumeResult(bch.ParametrizedSweep):
    """A class to represent the properties of a volume sample."""

    value = bch.ResultVar("ul", doc="The scalar value of the 3D volume field")
    occupancy = bch.ResultVar(
        "occupied", doc="If the value is > 0.5 this point is considered occupied"
    )
    interesting = bch.ResultVar("ul", doc="A more interesting scalar field")
    interesting_vec = bch.ResultVec(3, "vec", doc="A vector field with an interesting shape")
    interesting_vec_and_occ = bch.ResultVec(
        3, "vec", doc="The same vector field but only showing values in a sphere of radius 0.5"
    )


def bench_fn(point: VolumeSample) -> VolumeResult:
    """This function takes a 3D point as input and returns distance of that point to the origin.

    Args:
        point (VolumeSample): Sample point

    Returns:
        VolumeResult: Value at that point
    """
    output = VolumeResult()
    output.value = np.linalg.norm(np.array([point.x, point.y, point.z]))  # distance to origin
    output.occupancy = float(output.value < 0.5)
    # from https://plotly.com/python/3d-volume-plots/
    output.interesting = np.sin(np.pi * point.x) * np.cos(np.pi * point.z) * np.sin(np.pi * point.y)
    output.interesting_vec = [
        np.sin(np.pi * point.x),
        np.cos(np.pi * point.z),
        np.sin(np.pi * point.y),
    ]

    if output.occupancy:
        output.interesting_vec_and_occ = output.interesting_vec
    else:
        output.interesting_vec_and_occ = [0, 0, 0]

    return output


def example_floats3D(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """Example of how to perform a 3D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    bench = bch.Bench("Bencher_Example_Floats", bench_fn, VolumeSample)

    bench.plot_sweep(
        input_vars=[VolumeSample.param.x, VolumeSample.param.y, VolumeSample.param.z],
        result_vars=[
            VolumeResult.param.value,
            VolumeResult.param.occupancy,
            VolumeResult.param.interesting,
        ],
        title="Float 3D Example",
        description="""This example shows how to sample 3 floating point variables and plot a volumetric representation of the results.  The benchmark function returns the distance to the origin""",
        post_description="Here you can see concentric shells as the value of the function increases with distance from the origin. The occupancy graph should show a sphere with radius=0.5",
        run_cfg=run_cfg,
    )

    bench.plot_sweep(
        input_vars=[VolumeSample.param.x, VolumeSample.param.y, VolumeSample.param.z],
        result_vars=[
            VolumeResult.param.interesting_vec,
            VolumeResult.param.interesting_vec_and_occ,
        ],
        title="Float 3D Cone Example",
        description="""This example shows how to sample 3 floating point variables and plot a 3D vector field of the results.""",
        run_cfg=run_cfg,
    )

    return bench


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    example_floats3D(ex_run_cfg).show()



from typing import Protocol, Callable, List
import logging
from bencher.bench_cfg import BenchRunCfg, BenchCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.bencher import Bench

from copy import deepcopy


class Benchable(Protocol):
    def bench(self, run_cfg: BenchRunCfg) -> BenchCfg:
        ...


class BenchRunner:
    def __init__(
        self, bench_class=None, run_cfg: BenchRunCfg = BenchRunCfg(), publisher: Callable = None
    ) -> None:
        self.run_cfg = BenchRunner.setup_run_cfg(run_cfg)
        self.bench_fns = []
        self.publisher = publisher
        if bench_class is not None:
            self.add_bench(bench_class)

    @staticmethod
    def setup_run_cfg(run_cfg: BenchRunCfg = BenchRunCfg(), level: int = 1) -> BenchRunCfg:
        run_cfg_out = deepcopy(run_cfg)
        run_cfg_out.use_sample_cache = True
        run_cfg_out.only_hash_tag = True
        run_cfg_out.level = level
        return run_cfg_out

    @staticmethod
    def from_parametrized_sweep(
        class_instance: ParametrizedSweep, run_cfg: BenchRunCfg = BenchRunCfg()
    ):
        return Bench(f"bench_{class_instance.name}", class_instance, run_cfg=run_cfg)

    def add_run(self, bench_fn: Benchable) -> None:
        self.bench_fns.append(bench_fn)

    def add_bench(self, class_instance: ParametrizedSweep) -> None:
        def cb(run_cfg: BenchRunCfg) -> BenchCfg:
            bench = Bench(f"bench_{class_instance.name}", class_instance, run_cfg=run_cfg)
            return bench.plot_sweep(f"bench_{class_instance.name}")

        self.add_run(cb)

    def run(
        self,
        min_level: int = 2,
        max_level: int = 6,
        run_cfg: BenchRunCfg = None,
        publish: bool = False,
        debug: bool = True,
    ) -> List[BenchCfg]:
        results = []
        if run_cfg is not None:
            run_run_cfg = BenchRunner.setup_run_cfg(run_cfg)
        else:
            run_run_cfg = deepcopy(self.run_cfg)

        for lvl in range(min_level, max_level):
            for bch_fn in self.bench_fns:
                run_lvl = deepcopy(run_run_cfg)
                run_lvl.level = lvl
                logging.info(f"Running {bch_fn} at level: {lvl}")
                res = bch_fn(run_lvl)
                if publish and self.publisher is not None:
                    res.publish(remote_callback=self.publisher, debug=debug)
                results.append(res)
        return results

    # def show(self)
    # self.bench_fns[]



# Generated by CodiumAI
import unittest
from unittest.mock import Mock
from bencher.example.benchmark_data import SimpleBenchClass, AllSweepVars
import bencher as bch


class TestBenchRunner(unittest.TestCase):
    # Tests that bch.BenchRunner can be created with default configuration and the import statement in the bch.BenchRunner class is fixed
    def test_benchrunner_default_configuration_fixed(self):
        bench_runner = bch.BenchRunner()
        self.assertEqual(bench_runner.run_cfg.use_sample_cache, True)
        self.assertEqual(bench_runner.run_cfg.only_hash_tag, True)
        self.assertEqual(bench_runner.run_cfg.level, 1)
        self.assertEqual(bench_runner.publisher, None)
        self.assertEqual(bench_runner.bench_fns, [])

    # Tests that Benchable functions can be added to bch.BenchRunner instance
    def test_benchrunner_add_benchable_functions(self):
        bench_runner = bch.BenchRunner()
        bench_fn1 = Mock()
        bench_fn2 = Mock()
        bench_runner.add_run(bench_fn1)
        bench_runner.add_run(bench_fn2)
        self.assertEqual(len(bench_runner.bench_fns), 2)
        self.assertIn(bench_fn1, bench_runner.bench_fns)
        self.assertIn(bench_fn2, bench_runner.bench_fns)

    def test_benchrunner_handle_empty_list(self):
        bench_runner = bch.BenchRunner()
        results = bench_runner.run()
        self.assertEqual(len(results), 0)

    def test_benchrunner_benchable_class(self):
        bench_runner = bch.BenchRunner()
        bench_runner.add_bench(SimpleBenchClass())
        results = bench_runner.run(run_cfg=bch.BenchRunCfg(run_tag="1"))

        self.assertEqual(results[0].run_tag, "1")

    def test_benchrunner_benchable_class_run_constructor(self):
        bench_runner = bch.BenchRunner(run_cfg=bch.BenchRunCfg(run_tag="1"))
        bench_runner.add_bench(SimpleBenchClass())
        results = bench_runner.run()
        self.assertEqual(results[0].run_tag, "1")

    def test_benchrunner_level_1(self):
        results = bch.BenchRunner(AllSweepVars()).run(min_level=1)
        ds = results[0].ds.to_dataframe()
        self.assertEqual(ds.shape[0], 1)

    # # Tests that bch.BenchRunner can run Benchable functions with default configuration (fixed)
    # def test_benchrunner_run_default_configuration_fixed(self):

    #     bench_runner = bch.BenchRunner()
    #     bench_fn1 = Mock()
    #     bench_fn2 = Mock()
    #     bench_runner.add_run(bench_fn1)
    #     bench_runner.add_run(bench_fn2)
    #     results = bench_runner.run()

    #     self.assertEqual(len(results), 10)
    #     self.assertEqual(results[0].level, 1)
    #     self.assertEqual(results[1].level, 1)
    #     self.assertEqual(results[2].level, 2)
    #     self.assertEqual(results[3].level, 2)
    #     self.assertEqual(results[4].level, 3)
    #     self.assertEqual(results[5].level, 3)
    #     self.assertEqual(results[6].level, 4)
    #     self.assertEqual(results[7].level, 4)
    #     self.assertEqual(results[8].level, 5)
    #     self.assertEqual(results[9].level, 5)

    # Tests that bch.BenchRunner can run Benchable functions with custom configuration, after fixing the import statements
    # def test_benchrunner_run_custom_configuration_fixed_fixed_import_statements(self):

    #     bench_runner = bch.BenchRunner()
    #     bench_fn1 = Mock()
    #     bench_fn2 = Mock()
    #     bench_runner.add_run(bench_fn1)
    #     bench_runner.add_run(bench_fn2)
    #     run_cfg = bch.BenchRunCfg()
    #     run_cfg.use_sample_cache = False
    #     run_cfg.only_hash_tag = False
    #     run_cfg.level = 3
    #     results = bench_runner.run(run_cfg=run_cfg)
    #     self.assertEqual(len(results), 2)
    #     self.assertEqual(results[0].level, 3)
    #     self.assertEqual(results[1].level, 3)

    # Tests that bch.BenchRunner can publish results of Benchable functions (fixed)
    # def test_benchrunner_publish_results_fixed(self):
    #     class MockBenchable:
    #         def bench(self, run_cfg: bch.BenchRunCfg) -> bch.BenchCfg:
    #             return bch.BenchCfg()

    #     bench_runner = bch.BenchRunner(publisher=Mock())
    #     bench_fn1 = MockBenchable()
    #     bench_fn2 = MockBenchable()
    #     bench_runner.add_run(bench_fn1)
    #     bench_runner.add_run(bench_fn2)
    #     results = bench_runner.run(publish=True)
    #     self.assertEqual(len(results), 10)
    #     self.assertEqual(bench_runner.publisher.call_count, 10)
    #     self.assertEqual(bench_runner.publisher.call_args_list[0][0][0], results[0])
    #     self.assertEqual(bench_runner.publisher.call_args_list[1][0][0], results[1])
    #     self.assertEqual(bench_runner.publisher.call_args_list[2][0][0], results[2])
    #     self.assertEqual(bench_runner.publisher.call_args_list[3][0][0], results[3])
    #     self.assertEqual(bench_runner.publisher.call_args_list[4][0][0], results[4])
    #     self.assertEqual(bench_runner.publisher.call_args_list[5][0][0], results[5])
    #     self.assertEqual(bench_runner.publisher.call_args_list[6][0][0], results[6])
    #     self.assertEqual(bench_runner.publisher.call_args_list[7][0][0], results[7])
    #     self.assertEqual(bench_runner.publisher.call_args_list[8][0][0], results[8])
    #     self.assertEqual(bench_runner.publisher.call_args_list[9][0][0], results[9])

    # Tests that bch.BenchRunner can handle empty list of Benchable functions

    # Tests that bch.BenchRunner can handle empty list of Benchable functions
    # def test_benchrunner_handle_empty_list(self):

    #     def benchable(run_cfg:bch.BenchRunCfg)->bch.BenchCfg:
    #         bench = bch.Bench("sbc",SimpleBenchClass(),run_cfg=run_cfg)
    #         return bench.plot_sweep("sweep1")

    #     bench_runner = bch.BenchRunner()
    #     bench_runner.add_run(benchable)

    #     results = bench_runner.run(run_cfg=bch.BenchRunCfg(run_tag="1"))

    #     self.assertEqual(results[0].run_tag, "1")

    # Tests that bch.BenchRunner can handle empty list of Benchable functions



from enum import auto

from strenum import StrEnum


class PlotTypes(StrEnum):
    """A enum that contains all of the inbuilt plot types.  The enum is a strEnum so the value of the enum is a str representation of the name of itself.  e.g PlotTypes.swarmplot is equal to the string 'swarmplot'.  The PlotCollection class has an add() and remove() method that takes the string name of the plot to add or remove.  You can either pass raw string, or to help reduce typos and enhance discoverability use PlotTypes.xxxxxxxx."""

    # catplot enums
    swarmplot = auto()
    barplot = auto()
    violinplot = auto()
    boxenplot = auto()
    boxplot = auto()

    # table enums
    dataframe_multi_index = auto()
    dataframe_flat = auto()
    dataframe_mean = auto()
    xarray = auto()

    heatmap_2D = auto()
    heatmap_1D = auto()

    lineplot = auto()
    lineplot_hv = auto()
    lineplot_hv_overlay = auto()
    lineplot_hv_layout = auto()
    bar_hv = auto()
    # lineplot_hv_subplot = auto()
    # lineplot_hv_repeats = auto()

    # scatter2D_sns = auto()
    scatter_hv = auto()

    volume_plotly = auto()
    # cone_plotly = auto()
    surface_hv = auto()

    # hv_interactive = auto()



import os
import subprocess
import pytest
import unittest
import random
from shutil import rmtree
from copy import deepcopy
import logging

from hypothesis import given, settings, strategies as st

from datetime import datetime
from diskcache import Cache

from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function
from bencher import Bench, BenchCfg, BenchRunCfg, BenchPlotter


def get_hash_isolated_process() -> bytes:
    """This sets up bencher in a new process and prints a hash of the input config to the terminal which is then returned by this function.  The purpose is to set up bench from two different python process and make sure the hashes match"""
    result = subprocess.run(
        [
            "python3",
            "-c",
            "'from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut;import bencher as bch;cfg1 = bch.BenchCfg(input_vars=[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.noise_distribution],result_vars=[ExampleBenchCfgOut.param.out_sin],const_vars=[ExampleBenchCfgIn.param.noisy],repeats=5,over_time=False);print(cfg1.hash_persistent())'",
        ],
        stdout=subprocess.PIPE,
        check=False,
    )
    return result.stdout


def clear_autofig_folder() -> None:
    try:
        rmtree("autofig")
    except FileNotFoundError as e:
        logging.debug(e)
    os.mkdir("autofig")


# at the beginning of the test delete all the figures in the autofig folder.  At the end of the test they should be replaced with pixel perfect figures.  If the git repo is dirty at the end of the tests then CI will fail.
clear_autofig_folder()

# at the beginning of the tests clear the cache that checks for unique names
with Cache("unique_names") as c:
    c.clear()


# define a set of input variables that are used across multiple tests.  These configurations are
input_var_cat_permutations = [
    [ExampleBenchCfgIn.param.postprocess_fn],
    [ExampleBenchCfgIn.param.postprocess_fn, ExampleBenchCfgIn.param.noisy],
    [
        ExampleBenchCfgIn.param.postprocess_fn,
        ExampleBenchCfgIn.param.noisy,
        ExampleBenchCfgIn.param.noise_distribution,
    ],
]

# set up float variable permutations
input_var_float_permutations = [
    [ExampleBenchCfgIn.param.theta],
    [ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.postprocess_fn],
]

input_var_cat_float_permutations = input_var_cat_permutations + input_var_float_permutations

# Generating figures for the None case has some edge cases so make a separate variable to make it easier to debug
input_var_and_none_permutations = [] + input_var_cat_permutations

result_var_permutations = [
    [ExampleBenchCfgOut.param.out_sin],
    [ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos],
]


class TestBencher(unittest.TestCase):
    def create_bench(self) -> Bench:
        return Bench("test_bencher", bench_function, ExampleBenchCfgIn)

    @settings(deadline=10000)
    @given(
        input_vars=st.sampled_from(
            [
                [ExampleBenchCfgIn.param.theta],
                [ExampleBenchCfgIn.param.postprocess_fn],
                [ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.postprocess_fn],
            ]
        ),
        result_vars=st.sampled_from(result_var_permutations),
        const_vars=st.sampled_from([[]]),
        repeats=st.sampled_from([1, 2]),
        over_time=st.booleans(),
    )
    def test_bench_cfg_hash(self, input_vars, result_vars, const_vars, repeats, over_time):
        """check that identical inputs result in the same hash even if the object instances are not the same (this does not work by default with param and requires a custom hash function)"""

        cfg1 = BenchCfg(
            title="test_bencher_hash",
            input_vars=deepcopy(input_vars),
            result_vars=deepcopy(result_vars),
            const_vars=deepcopy(const_vars),
            run_cfg=BenchRunCfg(
                repeats=repeats,
                over_time=over_time,
                clear_history=True,  # should not affect hash
                auto_plot=False,
            ),
        )

        cfg2 = BenchCfg(
            title="test_bencher_hash",
            input_vars=deepcopy(input_vars),
            result_vars=deepcopy(result_vars),
            const_vars=deepcopy(const_vars),
            run_cfg=BenchRunCfg(
                repeats=repeats,
                over_time=over_time,
                clear_history=False,  # should not affect hash
                auto_plot=False,
            ),
        )

        self.assertEqual(
            cfg1.hash_persistent(include_repeats=True),
            cfg2.hash_persistent(include_repeats=True),
        )

    def test_bench_cfg_hash_isolated(self):
        """hash values only seem to not match if run in a separate process, so run the hash test in separate processes"""
        self.assertEqual(get_hash_isolated_process(), get_hash_isolated_process())

    # @pytest.mark.skip
    @settings(deadline=30000)
    @given(
        input_vars=st.sampled_from(input_var_cat_permutations),
        result_vars=st.sampled_from([[ExampleBenchCfgOut.param.out_sin]]),
    )
    def test_combinations_over_time(self, input_vars, result_vars) -> None:
        """check that up to 3 categorical values over time can be plotted"""
        # needed her instead of init because hypothesis calls this function multiple times after init() and the randomly generated data need to be the same each time to produce identical results to match the hand check plot iamges
        random.seed(42)
        bench = self.create_bench()
        for i in range(2):
            bench.plot_sweep(
                title="test_combinations_over_time",
                input_vars=input_vars,
                result_vars=result_vars,
                run_cfg=BenchRunCfg(
                    repeats=2,
                    over_time=True,
                    clear_history=i == 0,
                    print_pandas=False,
                    serve_panel=False,
                ),
                time_src=datetime(1970, 1, i + 1),  # repeatable time
            )

    # @pytest.mark.skip
    @settings(deadline=10000)
    @given(
        input_vars=st.sampled_from(input_var_cat_permutations),
        result_vars=st.sampled_from(
            [[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos]]
        ),
        repeats=st.sampled_from([20]),
        # repeats=st.sampled_from([1, 2]), #TODO this fails at the moment
    )
    def test_combinations(self, input_vars, result_vars, repeats) -> None:
        """check that up to 3 categorical and 1 float value without time can be plotted"""
        # needed her instead of init because hypothesis calls this function multiple times after init() and the randomly generated data need to be the same each time to produce identical results to match the hand check plot iamges
        random.seed(42)
        bench = self.create_bench()
        bench.plot_sweep(
            title="test_combinations",
            input_vars=input_vars,
            result_vars=result_vars,
            run_cfg=BenchRunCfg(
                repeats=repeats,
                print_pandas=False,
                over_time=False,
                serve_panel=False,
            ),
        )

    @settings(deadline=10000)
    @given(
        input_vars=st.sampled_from(
            [[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.offset]]
        ),
        result_vars=st.sampled_from(
            [[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos]]
        ),
        repeats=st.sampled_from([2]),
    )
    def test_pareto(self, input_vars, result_vars, repeats) -> None:
        """check that pareto optimisation works"""
        # needed her instead of init because hypothesis calls this function multiple times after init() and the randomly generated data need to be the same each time to produce identical results to match the hand check plot iamges
        random.seed(42)
        bench = self.create_bench()
        bench.plot_sweep(
            title="test_pareto_opt",
            input_vars=input_vars,
            result_vars=result_vars,
            run_cfg=BenchRunCfg(
                repeats=repeats,
                print_pandas=False,
                serve_panel=False,
                use_optuna=True,
                debug=True,
            ),
        )

    # TODO There are still name collisions when run on all possible inputs, but at the moment the name collisions end up plotting an identical graph anyway so it doesn't matter that much. Future work is to enable this test to confirm that all graph names are fully unique even if they have the same pixels.
    @pytest.mark.skip()
    @settings(deadline=10000)
    @given(
        input_vars=st.sampled_from(input_var_cat_permutations),
        result_vars=st.sampled_from(result_var_permutations),
        repeats=st.sampled_from([2]),
        # repeats=st.sampled_from([1, 2]), #TODO this fails at the moment
        over_time=st.booleans(),
    )
    def test_unique_file_names(self, input_vars, result_vars, repeats, over_time):
        """This tests that every single plot has a unique but meaningfull (not just a hash) name."""
        bench = self.create_bench()
        if over_time:
            for i in range(3):
                bench_cfg = bench.plot_sweep(
                    title="test_unique_filenames",
                    input_vars=input_vars,
                    result_vars=result_vars,
                    run_cfg=BenchRunCfg(
                        repeats=2,
                        over_time=True,
                        clear_history=i == 0,
                        print_pandas=False,
                        auto_plot=False,
                    ),
                    time_src=datetime(
                        1970, 1, i + 1
                    ),  # repeatable time so outputs are same at the pixel level
                )

        else:
            bench_cfg = bench.plot_sweep(
                title="test_unique_filenames",
                input_vars=input_vars,
                result_vars=result_vars,
                run_cfg=BenchRunCfg(
                    repeats=repeats,
                    over_time=False,
                    auto_plot=False,
                ),
            )

        bench_cfg.raise_duplicate_exception = False
        with Cache("unique_names") as name_cache:
            bench_repr = bench_cfg.__repr__()
            plots = BenchPlotter.plot(bench_cfg)
            for p in plots:
                if p.name is not None:
                    if p.name in name_cache:
                        self.fail(
                            f"this name already exists: \n\n\nA:{p.name}\n\n\nreprA:{bench_cfg.__repr__()}\n\n\nB:{name_cache[p.name]}",
                        )
                name_cache[p.name] = bench_cfg.__repr__()
            name_cache[bench_repr] = True

    @settings(deadline=10000)
    @given(
        over_time=st.booleans(),
    )
    def test_benching_cache_without_time(self, over_time) -> None:
        """check that the correct benching cache loads"""

        # set up inputs and results that are shared across runs
        title = "test_benching_cache"
        iv = [ExampleBenchCfgIn.param.theta]
        rv = [ExampleBenchCfgOut.param.out_sin]

        bench = self.create_bench()

        # run without caching and make sure any old caches are cleared
        bench.plot_sweep(
            title=title,
            input_vars=iv,
            result_vars=rv,
            run_cfg=BenchRunCfg(
                over_time=over_time, clear_cache=True, clear_history=True, auto_plot=False
            ),
        )

        self.assertEqual(bench.worker_wrapper_call_count, ExampleBenchCfgIn.param.theta.samples)

        bench2 = self.create_bench()
        # run again without caching, the function should be called again
        bench2.plot_sweep(
            title=title,
            input_vars=iv,
            result_vars=rv,
            run_cfg=BenchRunCfg(over_time=over_time, use_cache=False, auto_plot=False),
        )
        self.assertEqual(bench2.worker_wrapper_call_count, ExampleBenchCfgIn.param.theta.samples)

        # bench3 = self.create_bench()
        # run again with the cache turned on. The worker_wrapper_call_count should not increase because it loads cached results
        bench2.plot_sweep(
            title=title,
            input_vars=iv,
            result_vars=rv,
            run_cfg=BenchRunCfg(over_time=over_time, use_cache=True, auto_plot=False),
        )
        self.assertEqual(bench2.worker_wrapper_call_count, ExampleBenchCfgIn.param.theta.samples)

    @settings(deadline=10000)
    @given(noisy=st.booleans())
    def test_const_hashing(self, noisy) -> None:
        """check that const variables are hashed correctly. This test was created because setting a const variable was resulting in a hash value that changed over time even though the inputs were not changing.  The source of the problem was that the input config had a native param instead of a paramSweep object.  The native param objects don't have a constant hash because they include the .name field which changes for every instance of the param.  the paramSweep objects have the .name field removed from the hash so that hashes for the same inputs remain constant"""

        ExampleBenchCfgIn.param.theta.samples = 5

        logging.info(f"starting with const value noisy:{noisy}")

        bench = self.create_bench()

        # run without caching and make sure any old caches are cleared
        bench.plot_sweep(
            title="test_const_hashing",
            input_vars=[ExampleBenchCfgIn.param.theta],
            result_vars=[ExampleBenchCfgOut.param.out_sin],
            const_vars=[
                (ExampleBenchCfgIn.param.noisy, noisy),
            ],
            run_cfg=BenchRunCfg(clear_cache=True, clear_history=True, auto_plot=False),
        )
        self.assertEqual(
            bench.worker_wrapper_call_count,
            ExampleBenchCfgIn.param.theta.samples,
            "no cache used so the function should sample again",
        )
        logging.info("re-run and attempt to load from cache")

        bench2 = self.create_bench()
        # run again without caching, the function should be called again
        bench2.plot_sweep(
            title="test_const_hashing",
            input_vars=[ExampleBenchCfgIn.param.theta],
            result_vars=[ExampleBenchCfgOut.param.out_sin],
            const_vars=[
                (ExampleBenchCfgIn.param.noisy, noisy),
            ],
            run_cfg=BenchRunCfg(use_cache=True, auto_plot=False),
        )
        # the result should be cached so the call count should be the same as before
        self.assertEqual(
            bench2.worker_wrapper_call_count,
            0,
            "the worker should not be sampled as it should be loaded from the cache",
        )

    def test_forgetting_to_use_param(self) -> None:
        bench = self.create_bench()

        with self.assertRaises(TypeError):
            bench.plot_sweep(
                title="test_param_usage",
                input_vars=[ExampleBenchCfgIn.param.theta],
                result_vars=[ExampleBenchCfgOut.out_sin],  # forgot to use param here
            )

        with self.assertRaises(TypeError):
            bench.plot_sweep(
                title="test_param_usage",
                input_vars=[ExampleBenchCfgIn.theta],  # forgot to use param here
                result_vars=[ExampleBenchCfgOut.param.out_sin],
            )

        with self.assertRaises(TypeError):
            bench.plot_sweep(
                title="test_param_usage",
                input_vars=[ExampleBenchCfgIn.param.theta],
                result_vars=[ExampleBenchCfgOut.param.out_sin],
                const_vars=[(ExampleBenchCfgIn.offset, 0.1)],  # forgot to use param here
            )



# Generated by CodiumAI
from diskcache import Cache
import unittest
import bencher as bch

from bencher.example.benchmark_data import SimpleBenchClass


class TestBenchPlotServer(unittest.TestCase):
    # Tests that the plot server loads previously calculated benchmark data from the database
    def test_plot_server_load_data_from_database(self):
        sbc = SimpleBenchClass()
        bench = bch.Bench("test_bench_server", sbc)
        bench.plot_sweep(
            title="test_bench_server",
            input_vars=[sbc.param.var1],
            result_vars=[sbc.param.result],
            run_cfg=bch.BenchRunCfg(auto_plot=False),
        )
        bench.save()

        bps = bch.BenchPlotServer()

        bps.load_data_from_cache(bench.bench_name)

    # Tests that the plot server raises FileNotFoundError when no data is found in the cache
    def test_plot_server_no_data_in_cache(self):
        server = bch.BenchPlotServer()
        bench_name = "test_bench"
        with self.assertRaises(FileNotFoundError):
            server.load_data_from_cache(bench_name)

    # Tests that the plot server raises FileNotFoundError when no data is found in the database
    def test_plot_server_no_data_in_database(self):
        server = bch.BenchPlotServer()
        bench_name = "test_bench"
        with Cache("cachedir/benchmark_inputs") as cache:
            if bench_name in cache:
                with self.assertRaises(FileNotFoundError):
                    server.load_data_from_cache(bench_name)
            else:
                with self.assertRaises(FileNotFoundError):
                    server.load_data_from_cache(bench_name)



import unittest

# Generated by CodiumAI
from bencher.plotting.plot_library import PlotLibrary
from bencher.plotting.plot_types import PlotTypes


class TestPlotLibrary(unittest.TestCase):
    # Tests that the tables PlotCollection only contains table plots
    def test_tables_plot_collection(self) -> None:
        tables_plots = PlotLibrary.tables()
        self.assertIn(PlotTypes.dataframe_multi_index, tables_plots.plotters)
        self.assertIn(PlotTypes.dataframe_mean, tables_plots.plotters)
        self.assertEqual(len(tables_plots.plotters), 2)

    # Tests that the all PlotCollection contains all possible plots
    def test_all_plot_collection(self) -> None:
        all_plots = PlotLibrary.all()
        for pt in list(PlotTypes):
            self.assertIn(pt, all_plots.plotters)
        self.assertEqual(len(all_plots.plotters), len(PlotTypes))

    # Tests that the none PlotCollection contains no active plots
    def test_none_plot_collection(self) -> None:
        none_plots = PlotLibrary.none()
        self.assertEqual(len(none_plots.plotters), 0)

    def test_no_plotly_plot_collection(self) -> None:
        plt_col = PlotLibrary.with_plotly()
        self.assertIn(PlotTypes.surface_hv, plt_col.plotters)
        self.assertIn(PlotTypes.volume_plotly, plt_col.plotters)

    # Test that adding a plot that is not in the list of available plots raises a ValueError
    def test_adding_plot_not_in_available_plots(self) -> None:
        plt_col = PlotLibrary.none()
        with self.assertRaises(ValueError):
            plt_col.add("invalid_plot")

    # Test that removing a plot that is not in the list of active plots raises an Keyerror
    def test_removing_plot_not_in_active_plots(self) -> None:
        plt_col = PlotLibrary.tables()
        plt_col.remove(PlotTypes.dataframe_multi_index)
        self.assertEqual(len(plt_col.plotters), 1)  # 2 plots in tables by default and 1 removed
        with self.assertRaises(KeyError):
            plt_col.remove(PlotTypes.swarmplot)



# Generated by CodiumAI
from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.variables.inputs import BoolSweep
import param
import holoviews as hv
import unittest


class TestParametrizedSweep(unittest.TestCase):
    # Tests that the method returns an empty list when given a single input

    # Tests that the method returns a list of dimensions with computed values when compute_values=True
    def test_compute_values_true(self):
        sweep = ParametrizedSweep()
        sweep.param.one = param.Integer()
        sweep.param.two = param.Integer()
        dims = sweep.get_inputs_as_dims(compute_values=True)
        assert dims == []

    # Tests that the method returns a list of dimensions with all inputs when remove_dims=None
    def test_remove_dims_none(self):
        sweep = ParametrizedSweep()
        sweep.param.one = param.Integer()
        sweep.param.two = param.Integer()
        dims = sweep.get_inputs_as_dims(remove_dims=None)
        assert len(dims) == 0

    # Tests that the method returns an empty list when given empty inputs
    def test_empty_inputs(self):
        sweep = ParametrizedSweep()
        dims = sweep.get_inputs_as_dims()
        assert len(dims) == 0

    # Tests that the method returns a list of dimensions with all inputs when given a non-existent dimension name in remove_dims
    def test_remove_dims_nonexistent(self):
        sweep = ParametrizedSweep()
        sweep.param.one = param.Integer()
        sweep.param.two = param.Integer()
        dims = sweep.get_inputs_as_dims(remove_dims="three")
        assert len(dims) == 0

        # Tests that the method works when a callback function is provided

    def test_callback_provided(self):
        def callback(**kwargs):  # pylint: disable=unused-argument
            return {"hmap": hv.Curve([1, 2, 3])}

        p = ParametrizedSweep()
        p.call = callback
        dm = p.to_dynamic_map(callback=callback)
        hm = p.to_holomap(callback)
        assert isinstance(dm, hv.DynamicMap)
        assert isinstance(hm, hv.HoloMap)

    def test_no_callback_provided(self):
        p = ParametrizedSweep()
        dm = p.to_dynamic_map()
        assert isinstance(dm, hv.DynamicMap)

    def test_bool_default(self) -> None:
        class BoolDefaultFalse(ParametrizedSweep):
            bool_var = BoolSweep(default=False)

        instance = BoolDefaultFalse()
        self.assertFalse(instance.bool_var)

        class BoolDefaultTrue(ParametrizedSweep):
            bool_var = BoolSweep(default=True)

        instance = BoolDefaultTrue()
        self.assertTrue(instance.bool_var)



"""This file has some examples for how to perform basic benchmarking parameter sweeps"""
import bencher as bch

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


def example_1D_bool(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """This example shows how to sample a 1 dimensional categorical variable and plot the result of passing that parameter sweep to the benchmarking function

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """

    bencher = bch.Bench(
        "benchmarking_example_categorical1D",
        bench_function,
        ExampleBenchCfgIn,
    )

    # here we sample the input variable theta and plot the value of output1. The (noisy) function is sampled 20 times so you can see the distribution
    res = bencher.plot_sweep(
        title="Example 1D Categorical",
        input_vars=[ExampleBenchCfgIn.param.noisy],
        # result_vars=[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_bool],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        description=example_1D_bool.__doc__,
        run_cfg=run_cfg,
    )
    bencher.append(res.to_bar())

    return bencher


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.repeats = 3
    ex_run_cfg.print_pandas = True
    ex_run_cfg.over_time = False
    # ex_run_cfg.auto_plot = False

    b = example_1D_bool(ex_run_cfg)
    b.save()



import holoviews as hv
from mortgage import Loan
import bencher as bch


class Mortgage(bch.ParametrizedSweep):
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=5)
    principal = bch.FloatSweep(default=200000, bounds=(0, 1000000), units="$")
    interest = bch.FloatSweep(default=6, bounds=(1, 10), samples=5, units="%")
    term = bch.IntSweep(default=24, bounds=(1, 24), units="months")

    period = bch.IntSweep(default=12, bounds=(1, 24), units="month")

    # RESULT VARS
    installment_payment = bch.ResultVar(units="$")
    installment_interest = bch.ResultVar(units="$")
    installment_total_interest = bch.ResultVar(units="$")
    installment_balance = bch.ResultVar(units="$")
    installment_principal = bch.ResultVar(units="$")

    def call(self, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        loan = Loan(
            principal=self.principal,
            interest=self.interest / 100,
            term=self.term,
        )

        mort_value = []
        for i in self.param.period.values():
            installment = loan.schedule(i)
            self.installment_payment = float(installment.payment)
            self.installment_interest = float(installment.interest)
            self.installment_total_interest = float(installment.total_interest)
            self.installment_balance = float(installment.balance)
            self.installment_principal = float(installment.principal)

            mort_value.append((i, self.installment_balance))

        pt = hv.Curve(
            mort_value, self.param.period.as_dim(), self.param.installment_balance.as_dim()
        )
        return self.get_results_values_as_dict(pt)


if __name__ == "__main__":
    mort = Mortgage()
    bench = bch.Bench("Mortgage", mort.call, plot_lib=None)

    res = bench.plot_sweep(
        "Mortgage",
        [mort.param.interest],
        [
            mort.param.installment_balance,
            mort.param.installment_principal,
        ],
    )

    bench.append(res.to_curve())

    bench.append(res.to(hv.Table).opts(width=1000))

    bench.append(res.to_holomap())

    bench.append(res.to_grid())

    bench.append_tab(mort.to_dynamic_map(name="Mortgage Calculator"))

    bench.show()
    # bench.plot()



from enum import auto
from typing import List

import param
from param import Number
from strenum import StrEnum
import holoviews as hv
from bencher.utils import hash_sha1


class OptDir(StrEnum):
    minimize = auto()
    maximize = auto()
    none = auto()  # If none this var will not appear in pareto plots


class ResultVar(Number):
    """A class to represent result variables and the desired optimisation direction"""

    __slots__ = ["units", "direction"]

    def __init__(self, units="ul", direction: OptDir = OptDir.minimize, **params):
        Number.__init__(self, **params)
        self.units = units
        self.default = 0  # json is terrible and does not support nan values
        self.direction = direction

    def as_dim(self) -> hv.Dimension:
        return hv.Dimension((self.name, self.name), unit=self.units)

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1((self.units, self.direction))


class ResultVec(param.List):
    """A class to represent fixed size vector result variable"""

    __slots__ = ["units", "direction", "size"]

    def __init__(self, size, units="ul", direction: OptDir = OptDir.minimize, **params):
        param.List.__init__(self, **params)
        self.units = units
        self.default = 0  # json is terrible and does not support nan values
        self.direction = direction
        self.size = size

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1((self.units, self.direction))

    def index_name(self, idx: int) -> str:
        """given the index of the vector, return the column name that

        Args:
            idx (int): index of the result vector

        Returns:
            str: column name of the vector for the xarray dataset
        """

        mapping = ["x", "y", "z"]
        if idx < 3:
            index = mapping[idx]
        else:
            index = idx
        return f"{self.name}_{index}"

    def index_names(self) -> List[str]:
        """Returns a list of all the xarray column names for the result vector

        Returns:
            list[str]: column names
        """
        return [self.index_name(i) for i in range(self.size)]


class ResultHmap(param.Parameter):
    """A class to represent a holomap return type"""

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return hash_sha1(self)


# class ResultList(param.Parameter):
#     """A class to unknown size vector result variable"""

#     __slots__ = ["units", "dim_name", "dim_units", "indices"]

#     def __init__(
#         self,
#         index_name: str,
#         index_units: str,
#         default=ResultSeries(),
#         units="ul",
#         indices: List[float] = None,
#         instantiate=True,
#         **params,
#     ):
#         param.Parameter.__init__(self, default=default, instantiate=instantiate, **params)
#         self.units = units
#         self.dim_name = index_name
#         self.dim_units = index_units
#         self.indices = indices

#     def hash_persistent(self) -> str:
#         """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
#         return hash_sha1((self.units, self.dim_name, self.dim_units))

# class ResultSeries:
#     """A class to represent a vector of results, it also includes an index similar to pandas.series"""

#     def __init__(self, values=None, index=None) -> None:
#         self.values = []
#         self.index = []
#         self.set_index_and_values(values, index)

#     def append(self, value: float | int, index: float | int | str = None) -> None:
#         """Add a value and index to the result series

#         Args:
#             value (float | int): result value of the series
#             index (float | int | str, optional): index value of series, the same as a pandas.series index  If no value is passed an integer index is automatically created. Defaults to None.
#         """

#         if index is None:
#             self.index.append(len(self.values))
#         else:
#             self.index.append(index)
#         self.values.append(value)

#     def set_index_and_values(
#         self, values: List[float | int], index: List[float | int | str] = None
#     ) -> None:
#         """Add values and indices to the result series

#         Args:
#             value (List[float | int]): result value of the series
#             index (List[float | int | str], optional): index value of series, the same as a pandas.series index  If no value is passed an integer index is automatically created. Defaults to None.
#         """
#         if values is not None:
#             if index is None:
#                 self.index = list(range(len(values)))
#             else:
#                 self.index = index
#             self.values = values



from typing import Optional

import panel as pn
import logging
import xarray as xr

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange, PltCntCfg
from bencher.plt_cfg import PltCfgBase
from bencher.plotting.plot_types import PlotTypes
from bencher.bench_cfg import BenchCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep

from bencher.plotting_functions import wrap_long_time_labels
import holoviews as hv
import plotly.graph_objs as go


from bencher.variables.results import ResultVar


def plot_float_cnt_2(plt_cnt_cfg: PltCntCfg, rv: ResultVar, debug: bool) -> PltCfgBase:
    """A function for determining the plot settings if there are 2 float variable and updates the PltCfgBase

    Args:
        sns_cfg (PltCfgBase): See PltCfgBase definition
        plt_cnt_cfg (PltCntCfg): See PltCntCfg definition

    Returns:
        PltCfgBase: See PltCfgBase definition
    """
    xr_cfg = PltCfgBase()
    if plt_cnt_cfg.float_cnt == 2:
        logging.info(f"surface plot: {rv.name}")
        xr_cfg.plot_callback_xra = xr.plot.plot
        xr_cfg.x = plt_cnt_cfg.float_vars[0].name
        xr_cfg.y = plt_cnt_cfg.float_vars[1].name
        xr_cfg.xlabel = f"{xr_cfg.x} [{plt_cnt_cfg.float_vars[0].units}]"
        xr_cfg.ylabel = f"{xr_cfg.y} [{plt_cnt_cfg.float_vars[1].units}]"
        xr_cfg.zlabel = f"{rv.name} [{rv.units}]"
        xr_cfg.title = f"{rv.name} vs ({xr_cfg.x} and {xr_cfg.y})"

        if plt_cnt_cfg.cat_cnt >= 1:
            logging.info("surface plot with 1 categorical")
            xr_cfg.row = plt_cnt_cfg.cat_vars[0].name
            xr_cfg.num_rows = len(plt_cnt_cfg.cat_vars[0].values(debug))
        if plt_cnt_cfg.cat_cnt >= 2:
            logging.info("surface plot with 2> categorical")
            xr_cfg.col = plt_cnt_cfg.cat_vars[1].name
            xr_cfg.num_cols = len(plt_cnt_cfg.cat_vars[1].values(debug))
    return xr_cfg


def plot_surface_plotly(
    bench_cfg: BenchCfg, rv: ParametrizedSweep, xr_cfg: PltCfgBase
) -> pn.pane.Plotly:
    """Given a benchCfg generate a 2D surface plot
    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot
        xr_cfg (PltCfgBase): config of x,y variables
    Returns:
        pn.pane.Plotly: A 2d surface plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    da = bench_cfg.ds[rv.name].transpose()

    mean = da.mean("repeat")

    x = da.coords[xr_cfg.x]
    y = da.coords[xr_cfg.y]

    opacity = 0.3

    surfaces = [go.Surface(x=x, y=y, z=mean)]

    if bench_cfg.repeats > 1:
        std_dev = da.std("repeat")
        surfaces.append(go.Surface(x=x, y=y, z=mean + std_dev, showscale=False, opacity=opacity))
        surfaces.append(go.Surface(x=x, y=y, z=mean - std_dev, showscale=False, opacity=opacity))

    eye_dis = 1.7
    layout = go.Layout(
        title=xr_cfg.title,
        width=700,
        height=700,
        scene=dict(
            xaxis_title=xr_cfg.xlabel,
            yaxis_title=xr_cfg.ylabel,
            zaxis_title=xr_cfg.zlabel,
            camera={"eye": {"x": eye_dis, "y": eye_dis, "z": eye_dis}},
        ),
    )

    fig = {"data": surfaces, "layout": layout}

    return pn.pane.Plotly(fig)


class SurfacePlot:
    def surface_hv(self, pl_in: PlotInput) -> Optional[pn.panel]:
        """Given a benchCfg generate a 2D surface plot

        Args:
            bench_cfg (BenchCfg): description of benchmark
            rv (ParametrizedSweep): result variable to plot
            xr_cfg (PltCfgBase): config of x,y variables

        Returns:
            pn.pane.holoview: A 2d surface plot as a holoview in a pane
        """
        if PlotFilter(
            float_range=VarRange(2, 2),
            cat_range=VarRange(0, None),
            vector_len=VarRange(1, 1),
            result_vars=VarRange(1, 1),
        ).matches(pl_in.plt_cnt_cfg):
            xr_cfg = plot_float_cnt_2(pl_in.plt_cnt_cfg, pl_in.rv, pl_in.bench_cfg.debug)
            bench_cfg = pl_in.bench_cfg
            rv = pl_in.rv

            bench_cfg = wrap_long_time_labels(bench_cfg)

            alpha = 0.3

            da = bench_cfg.ds[rv.name]

            print(da.to_dataframe())
            mean = da.mean("repeat")
            print(mean.to_dataframe())

            # TODO a warning suggests setting this parameter, but it does not seem to help as expected, leaving here to fix in the future
            # hv.config.image_rtol = 1.0

            ds = hv.Dataset(mean)
            print(ds)

            try:
                surface = ds.to(hv.Surface, vdims=[pl_in.rv.name])
            except Exception:
                return plot_surface_plotly(bench_cfg, rv, xr_cfg)

            try:
                surface = surface.opts(colorbar=True)
            except Exception as e:
                logging.warning(e)

            if bench_cfg.repeats > 1:
                std_dev = da.std("repeat")
                surface *= (
                    hv.Dataset(mean + std_dev)
                    .to(hv.Surface)
                    .opts(alpha=alpha, colorbar=False, backend="plotly")
                )
                surface *= (
                    hv.Dataset(mean - std_dev)
                    .to(hv.Surface)
                    .opts(alpha=alpha, colorbar=False, backend="plotly")
                )

            surface = surface.opts(
                width=800,
                height=800,
                zlabel=xr_cfg.zlabel,
                title=xr_cfg.title,
                backend="plotly",
            )

            if bench_cfg.render_plotly:
                hv.extension("plotly")
                out = surface
            else:
                # using render disabled the holoviews sliders :(
                out = hv.render(surface, backend="plotly")
            return pn.Column(out, name=PlotTypes.surface_hv)

        return None



import holoviews as hv
import panel as pn
from mortgage import Loan
import bencher as bch
from example_investment import Investment


class EconomicClimate(bch.ParametrizedSweep):
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=5)


class Mortgage(EconomicClimate):
    # INPUTS
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=5)

    principal = bch.FloatSweep(default=200000, bounds=(0, 1000000), units="$")
    interest = bch.FloatSweep(default=6, bounds=(1, 10), samples=5, units="%")
    period = bch.IntSweep(default=24, bounds=(1, 30 * 12), units="months")

    installment_num = bch.IntSweep(default=1, bounds=(1, 24), units="month")

    # RESULT VARS
    installment_payment = bch.ResultVar(units="$")
    installment_interest = bch.ResultVar(units="$")
    installment_total_interest = bch.ResultVar(units="$")
    installment_balance = bch.ResultVar(units="$")
    installment_principal = bch.ResultVar(units="$")

    def calc_loan(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)
        loan = Loan(
            principal=self.principal,
            interest=self.interest / 100,
            term=self.period,
        )

        installment = loan.schedule(self.installment_num)

        self.installment_payment = installment.payment
        self.installment_interest = installment.interest
        self.installment_total_interest = installment.total_interest
        self.installment_balance = installment.balance
        self.installment_principal = installment.principal

        return self.get_results_values_as_dict()

    def call(self, **kwargs):
        return self.calc_loan(**kwargs)


# res = bench.plot_sweep(
#     "Mortage",
#     input_vars=[mort_sim.param.installment_num],
#     result_vars=[
#         mort_sim.param.installment_payment,
#         mort_sim.param.installment_interest,
#         mort_sim.param.installment_total_interest,
#         mort_sim.param.installment_balance,
#         mort_sim.param.installment_principal,
#     ],
#     plot_lib=bch.PlotLibrary.none(),
# )

# res = bench.plot_sweep(
#     "Mortage vs Interest Rates",
#     input_vars=[mort_sim.param.installment_num, mort_sim.param.interest],
#     result_vars=[
#         # mort_sim.param.installment_payment,
#         # mort_sim.param.installment_interest,
#         mort_sim.param.installment_total_interest,
#         mort_sim.param.installment_balance,
#         mort_sim.param.installment_principal,
#     ],
# )


class PersonalFinances(Mortgage):
    inflation = bch.FloatSweep(default=8, bounds=[0.5, 15], units="%", samples=4)

    net_income = bch.FloatSweep(default=3000, bounds=(0, 5000))
    savings = bch.FloatSweep(default=1000, bounds=(0, 3000))
    mortgage_payment = bch.FloatSweep(default=1000, bounds=(500, 2000))

    monthly_contribution = bch.FloatSweep(default=0, bounds=[0, 1000], units="$", samples=4)
    investment_interest = bch.FloatSweep(default=6.0, bounds=(0.01, 10), units="%", samples=4)
    tax_rate = bch.IntSweep(default=0, bounds=(0, 40), samples=2)
    period = bch.IntSweep(default=12, bounds=[1, 12], units="month")
    starting_value = bch.FloatSweep(default=1000, bounds=(0, 2000), samples=3)

    discretionary_income = bch.FloatSweep(default=500, bounds=(0, 1000), units="$")
    investment_to_mortgage_ratio = bch.FloatSweep(default=0.3, bounds=(0, 1), units="ratio")
    investments_value = bch.ResultVar(default=0, units="$")

    # results
    net_wealth = bch.ResultVar(default=0, units="$")
    investments_value = bch.ResultVar(default=0, units="$")

    def call(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)

        loan = Loan(
            principal=self.principal,
            interest=self.investment_interest / 100,
            term=self.period,
        )

        installment = loan.schedule(self.installment_num)

        self.installment_payment = installment.payment
        self.installment_interest = installment.interest
        self.installment_total_interest = installment.total_interest
        self.installment_balance = installment.balance
        self.installment_principal = float(installment.principal)

        # print(self.installment_payment)
        self.investments_value = self.starting_value
        values = []

        # mort_inv = self.net_income - self.discretionary_income
        # self.monthly_contribution = (
        #     self.net_income
        #     - self.discretionary_income
        #     # - self.installment_payment
        # )
        # self.mortgage_payment
        # self.monthly_contribution = (
        #     self.post_tax_income * self.investment_to_mortgage_ratio
        # )

        for i in range(self.period):
            self.investments_value += self.monthly_contribution
            new_interest = self.investments_value * (self.investment_interest / 100.0)
            self.investments_value += new_interest * (1 + self.tax_rate)
            self.investments_value *= 1 - (self.inflation / 100.0)
            values.append((i, self.investments_value))

        # self.calc_loan(**kwargs)

        self.net_wealth = self.investments_value + self.installment_principal
        print(values)
        return self.get_results_values_as_dict(
            hv.Curve(values)
            # hv.Curve(values)
        )

    def call_vec(self, **kwargs):
        return self.call(**kwargs)


mort_sim = Mortgage()
inv = Investment()
pf = PersonalFinances()

bench = bch.Bench("Mortgage", mort_sim.calc_loan)


pn.Row(inv.to_dynamic_map(inv.call)).show()


bench.plot_sweep("lol")

bench.worker = inv.call

res = bench.plot_sweep(
    title="Investment Calculator",
    description="This calculator lets you calculate the value of your investment over time",
    input_vars=[
        # inv.param.period,
        inv.param.interest,
        inv.param.monthly_contribution,
        inv.param.inflation,
    ],
    # const_vars=[(inv.param.interest, .06)],
    result_vars=[inv.param.portfolio_value],
    # plot_lib=bch.PlotLibrary.default().add(bch.PlotTypes.lineplot_hv),
)

bench.append(res.to_grid())
# grid(["inflation", "monthly_contribution"]))
bench.append_tab(inv.to_dynamic_map(inv.call))

# bench.append_tab(inv.to_holomap(inv.call).layout(kdims=["inflation"]))


# bench.append_tab(pf.to_dynamic_map(pf.call_vec))

bench.show()

# bench.worker = inv.call
bench.worker = pf.call

res = bench.plot_sweep(
    title="Mortgage Simulator",
    description="This calculator lets you calculate the value of your investment over time",
    input_vars=[
        pf.param.period,
        pf.param.investment_interest,
        inv.param.monthly_contribution,
        inv.param.inflation,
    ],
    # const_vars=[(inv.param.interest, .06)],
    result_vars=[pf.param.investments_value],
    # plot_lib=bch.PlotLibrary.default().add(bch.PlotTypes.lineplot_hv),
    plot_lib=bch.PlotLibrary.none(),
)

# bench.append(res.to_grid().overlay("month"))

bench.append(res.to_curve().grid(["inflation", "monthly_contribution"]).opts())


# res = bench.plot_sweep(
#     title="Investment Calculator",
#     description="This calculator lets you calculate the value of your investment over time",
#     input_vars=[
#         inv.param.period,
#         # inv.param.interest,
#         # inv.param.monthly_contribution,
#         # inv.param.inflation,
#     ],
#     # const_vars=[(inv.param.interest, .06)],
#     result_vars=[inv.param.portfolio_value],
#     # plot_lib=bch.PlotLibrary.default().add(bch.PlotTypes.lineplot_hv),
# )


# print(inv.get_inputs_as_dims())

# bench.plots_instance.append(
#     hv.DynamicMap(callback=inv.plot, kdims=inv.get_inputs_as_dims()).opts(
#         width=1000, height=1000
#     )
# )


# bench.plots_instance.append(
#     hv.DynamicMap(Investment(), Investment().param.get_inputs_as_dims(True))
# ).opts(width=1000, height=1000)

# bench.append(res.to_curve().opts(width=1000, height=1000))


# bench.append(res.to_curve().grid(["inflation", "monthly_contribution"]).opts())


# bench.

# bench.append(res.to_curve().grid())
# bench.append(res.to_grid())


# ds = hv.Dataset(res.ds)
# # pt = ds.to(hv.Curve,vdims=vdims)

# bench.append(
#     ds.to(
#         hv.Curve, vdims=["installment_balance", "installment_total_interest"]
#     ).layout("interest")
# )

# bench.append(
#     ds.to(hv.Curve, vdims=["installment_total_interest"]).layout("interest")
# )

# bench.append(
#     ds.to(hv.Curve, vdims=["installment_total_interest"]).overlay("interest")
# )

# bench.append(res.to_curve(vdims=["installment_balance"]).overlay("interest"))

# bench.plots_instance.append(res.to_nd_layout())

bench.show()



import unittest

from bencher.example.benchmark_data import AllSweepVars


# def get_sweep_hash_isolated_process() -> str:
#     """get has values from a separate process as by default hashes across process are not the same"""
#     os.system(
#         "python3 -c 'from bencher.example.benchmark_data import AllSweepVars;ex = AllSweepVars();print(ex.__repr__());print(ex.hash_persistent())' > hashed_vars_comparison_tmp"
#     )

#     with open("hashed_vars_comparison_tmp", "r", encoding="utf-8") as myfile:
#         res = myfile.read()
#     os.remove("hashed_vars_comparison_tmp")
#     return res


class TestBencherHashing(unittest.TestCase):
    def test_sweep_hashes(self) -> None:
        """check that separate instances with identical data have the same hash"""
        ex = AllSweepVars()
        ex2 = AllSweepVars()

        self.assertEqual(
            ex.param.var_float.hash_persistent(), ex2.param.var_float.hash_persistent()
        )
        self.assertEqual(ex.param.var_int.hash_persistent(), ex2.param.var_int.hash_persistent())
        self.assertEqual(ex.param.var_enum.hash_persistent(), ex2.param.var_enum.hash_persistent())

        print(ex.__repr__())
        print(ex2.__repr__())

        self.assertEqual(ex.hash_persistent(), ex2.hash_persistent())

    def test_hash_sweep(self) -> None:
        """hash values only seem to not match if run in a separate process, so run the hash test in separate processes"""

        asv = AllSweepVars()
        asv2 = AllSweepVars()

        self.assertEqual(
            asv.hash_persistent(),
            asv2.hash_persistent(),
            "The classes should have equal hash when it has identical values",
        )

        asv2.var_float = 1
        self.assertNotEqual(
            asv.hash_persistent(),
            asv2.hash_persistent(),
            "The classes should not have equal hash when they have different values",
        )

    # turn this off as its quite expensive and the new hash function is reliable, turn back on for release
    # def test_hash_sweep_isolated(self) -> None:
    #     """hash values only seem to not match if run in a separate process, so run the hash test in separate processes"""

    #     self.assertNotEqual(
    #         len(get_sweep_hash_isolated_process()), 0, "make sure the hash is getting returned"
    #     )

    #     self.assertEqual(
    #         get_sweep_hash_isolated_process(),
    #         get_sweep_hash_isolated_process(),
    #         "make sure the hashes are equal",
    #     )


# print(get_sweep_hash_isolated_process())



import bencher as bch
import numpy as np
import optuna
from optuna.samplers import TPESampler


def objective(trial):
    x = trial.suggest_float("x", -10, 10)
    return x**2


study = optuna.create_study(sampler=TPESampler())
study.optimize(objective, n_trials=10)

rast_range = 1.5

optimal_value = 0.1234


class ToyOptimisationProblem(bch.ParametrizedSweep):
    input1 = bch.FloatSweep(default=0, bounds=[-rast_range, rast_range], samples=10)
    input2 = bch.FloatSweep(default=0, bounds=[-rast_range, rast_range], samples=10)
    offset = bch.FloatSweep(default=0, bounds=[-rast_range, rast_range])

    bump_scale = bch.FloatSweep(default=1.5, bounds=[1, 10])

    # RESULTS
    output = bch.ResultVar(bch.OptDir.minimize)

    def rastrigin(self, **kwargs) -> dict:
        """A modified version of the rastrigin function which is very difficult to find the global optimum
        https://en.wikipedia.org/wiki/Rastrigin_function

        Returns:
            dict: dictionary of return values
        """
        self.update_params_from_kwargs(**kwargs)
        x = np.array([self.input1 + optimal_value, self.input2 + optimal_value])

        self.output = (
            np.sum(x * x - self.bump_scale * np.cos(self.bump_scale * np.pi * x))
            + self.bump_scale * np.size(x)
            + self.offset
        )
        return self.get_results_values_as_dict()


def optuna_rastrigin(run_cfg=bch.BenchRunCfg()):
    explorer = ToyOptimisationProblem()

    bench = bch.Bench("Rastrigin", explorer.rastrigin)

    bench.to_optuna(
        input_vars=[explorer.param.input1, explorer.param.input2],
        result_vars=[explorer.param.output],
    )

    bench.get_panel(False).append(
        f"The optimal value should be input1:{-optimal_value},input2:{-optimal_value} with a value of 0"
    )

    run_cfg.use_optuna = True
    res = bench.plot_sweep(
        "Rastrigin",
        input_vars=[explorer.param.input1, explorer.param.input2],
        result_vars=[explorer.param.output],
        run_cfg=run_cfg,
    )

    bench.to_optuna_from_sweep(res, 10)

    return bench


if __name__ == "__main__":
    optuna_rastrigin().show()



from __future__ import annotations

from bencher.plotting.plot_collection import PlotCollection
from bencher.plotting.plot_types import PlotTypes
from bencher.plotting.plots.catplot import Catplot
from bencher.plotting.plots.heatmap import Heatmap
from bencher.plotting.plots.lineplot import Lineplot
from bencher.plotting.plots.tables import Tables
from bencher.plotting.plots.volume import VolumePlot
from bencher.plotting.plots.surface import SurfacePlot

# from bencher.plotting.plots.scatterplot import Scatter


from bencher.plotting.plots.hv_interactive import HvInteractive


class PlotLibrary:
    """This class is a collection of static factory methods that define sets of plots.  It provides a sensible default PlotCollection out of the box but also enables the creation of custom PlotCollections"""

    @staticmethod
    def setup_sources() -> PlotCollection:
        """Shared method that collects all the possible plot types.  This is used by all the other methods that select different subsections of the plots

        Returns:
            PlotCollection: A PlotCollection with sources of plots defined, but none selected
        """
        plt_col = PlotCollection()
        plt_col.add_plotter_source(Catplot())
        plt_col.add_plotter_source(Tables())
        plt_col.add_plotter_source(Heatmap())
        plt_col.add_plotter_source(Lineplot())
        # plt_col.add_plotter_source(Scatter())
        plt_col.add_plotter_source(HvInteractive())
        plt_col.add_plotter_source(VolumePlot())
        plt_col.add_plotter_source(SurfacePlot())

        return plt_col

    @staticmethod
    def default() -> PlotCollection:
        """Default PlotCollection.  This set of plots should provide a reasonable set of plots that cover the most common data plotting use cases

        Returns:
            PlotCollection: sensible default for PlotCollection
        """
        plt_col = PlotLibrary.setup_sources()
        # plt_col.add(PlotTypes.scatter_hv)

        plt_col.add(PlotTypes.swarmplot)
        plt_col.add(PlotTypes.heatmap_2D)  # 2d image of a surface
        # plt_col.add(PlotTypes.lineplot)
        plt_col.add(PlotTypes.lineplot_hv)
        plt_col.add(PlotTypes.lineplot_hv_overlay)
        plt_col.add(PlotTypes.lineplot_hv_layout)
        plt_col.add(PlotTypes.scatter_hv)
        # plt_col.add(PlotTypes.bar_hv)
        # plt_col.add(PlotTypes.cone_plotly)
        # plt_col.add(PlotTypes.lineplot_hv_subplot)
        # plt_col.add(PlotTypes.scatter2D_sns)
        # plt_col.add(PlotTypes.hv_interactive)

        return plt_col

    @staticmethod
    def with_plotly() -> PlotCollection:
        """Display only table summaries of the data

        Returns:
            PlotCollection: Only table plots
        """
        defaults = PlotLibrary.default()
        return defaults.add(PlotTypes.surface_hv).add(PlotTypes.volume_plotly)

    @staticmethod
    def tables() -> PlotCollection:
        """Display only table summaries of the data

        Returns:
            PlotCollection: Only table plots
        """
        plt_col = PlotLibrary.setup_sources()
        plt_col.add(PlotTypes.dataframe_multi_index)
        plt_col.add(PlotTypes.dataframe_mean)
        return plt_col

    @staticmethod
    def all() -> PlotCollection:
        """Display the data in every possible way. All plots are included

        Returns:
            PlotCollection: All possible plots
        """
        plt_col = PlotLibrary.setup_sources()
        for pt in PlotTypes:  # iterate through all plot enums
            plt_col.add(pt)
        return plt_col

    @staticmethod
    def none() -> PlotCollection:
        """Set up a plot collection with no active plots

        Returns:
            PlotCollection: No plots
        """
        return PlotLibrary.setup_sources()



from typing import Optional

import panel as pn
import plotly.graph_objs as go
import logging
import xarray as xr

from bencher.bench_cfg import BenchCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep

from bencher.plotting.plot_filter import PlotFilter, PlotInput, VarRange, PltCntCfg
from bencher.plt_cfg import PltCfgBase
from bencher.plotting.plot_types import PlotTypes

from bencher.plotting_functions import wrap_long_time_labels


def plot_float_cnt_3(sns_cfg: PltCfgBase, plt_cnt_cfg: PltCntCfg, debug: bool) -> PltCfgBase:
    """A function for determining the plot settings if there are 2 float variable and updates the PltCfgBase

    Args:
        sns_cfg (PltCfgBase): See PltCfgBase definition
        plt_cnt_cfg (PltCntCfg): See PltCntCfg definition

    Returns:
        PltCfgBase: See PltCfgBase definition
    """
    xr_cfg = PltCfgBase(**sns_cfg.as_dict())

    if plt_cnt_cfg.float_cnt >= 3:
        logging.info("volume plot")
        sns_cfg.plot_callback = None  # all further plots are surfaces
        xr_cfg.plot_callback_xra = xr.plot.plot
        xr_cfg.x = plt_cnt_cfg.float_vars[0].name
        xr_cfg.y = plt_cnt_cfg.float_vars[1].name
        xr_cfg.z = plt_cnt_cfg.float_vars[2].name
        xr_cfg.xlabel = f"{xr_cfg.x} [{plt_cnt_cfg.float_vars[0].units}]"
        xr_cfg.ylabel = f"{xr_cfg.y} [{plt_cnt_cfg.float_vars[1].units}]"
        xr_cfg.zlabel = f"{xr_cfg.z} [{plt_cnt_cfg.float_vars[2].units}]"
        if plt_cnt_cfg.cat_cnt >= 1:
            logging.info("volume plot with 1 categorical")
            xr_cfg.row = plt_cnt_cfg.cat_vars[0].name
            xr_cfg.num_rows = len(plt_cnt_cfg.cat_vars[0].values(debug))
        if plt_cnt_cfg.cat_cnt >= 2:
            logging.info("volume plot with 2> categorical")
            xr_cfg.col = plt_cnt_cfg.cat_vars[1].name
            xr_cfg.num_cols = len(plt_cnt_cfg.cat_vars[1].values(debug))
    return xr_cfg


def plot_volume_plotly(
    bench_cfg: BenchCfg, rv: ParametrizedSweep, xr_cfg: PltCfgBase
) -> pn.pane.Plotly:
    """Given a benchCfg generate a 3D surface plot

    Args:
        bench_cfg (BenchCfg): description of benchmark
        rv (ParametrizedSweep): result variable to plot
        xr_cfg (PltCfgBase): config of x,y variables

    Returns:
        pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    """

    bench_cfg = wrap_long_time_labels(bench_cfg)

    da = bench_cfg.ds[rv.name]

    mean = da.mean("repeat")

    opacity = 0.1

    meandf = mean.to_dataframe().reset_index()

    data = [
        go.Volume(
            x=meandf[xr_cfg.x],
            y=meandf[xr_cfg.y],
            z=meandf[xr_cfg.z],
            value=meandf[rv.name],
            isomin=meandf[rv.name].min(),
            isomax=meandf[rv.name].max(),
            opacity=opacity,
            surface_count=20,
        )
    ]

    layout = go.Layout(
        title=f"{rv.name} vs ({xr_cfg.x} vs {xr_cfg.y} vs {xr_cfg.z})",
        width=700,
        height=700,
        margin=dict(t=50, b=50, r=50, l=50),
        scene=dict(
            xaxis_title=xr_cfg.xlabel,
            yaxis_title=xr_cfg.ylabel,
            zaxis_title=xr_cfg.zlabel,
        ),
    )

    fig = dict(data=data, layout=layout)

    return pn.pane.Plotly(fig, name=PlotTypes.volume_plotly)


# def plot_cone_plotly(
#     bench_cfg: BenchCfg, rv: ParametrizedSweep, xr_cfg: PltCfgBase
# ) -> pn.pane.Plotly:
#     """Given a benchCfg generate a 3D surface plot

#     Args:
#         bench_cfg (BenchCfg): description of benchmark
#         rv (ParametrizedSweep): result variable to plot
#         xr_cfg (PltCfgBase): config of x,y variables

#     Returns:
#         pn.pane.Plotly: A 3d volume plot as a holoview in a pane
#     """

#     bench_cfg = wrap_long_time_labels(bench_cfg)

#     names = rv.index_names()

#     # da = bench_cfg.ds[names[0]]

#     opacity = 1.0

#     df = bench_cfg.get_dataframe()

#     print("size before removing zero size vectors", df.shape)
#     df = df.loc[(df[names[0]] != 0.0) | (df[names[1]] != 0.0) | (df[names[2]] != 0.0)]
#     print("size after removing zero size vectors", df.shape)

#     data = [
#         go.Cone(
#             x=df[xr_cfg.x],
#             y=df[xr_cfg.y],
#             z=df[xr_cfg.z],
#             u=df[names[0]],
#             v=df[names[1]],
#             w=df[names[2]],
#             # sizemode="absolute",
#             # sizeref=2,
#             anchor="tail",
#             opacity=opacity,
#         )
#     ]

#     layout = go.Layout(
#         title=f"{rv.name} vs ({xr_cfg.x} vs {xr_cfg.y} vs {xr_cfg.z})",
#         autosize=True,
#         width=700,
#         height=700,
#         margin=dict(t=50, b=50, r=50, l=50),
#         scene=dict(
#             xaxis_title=xr_cfg.xlabel,
#             yaxis_title=xr_cfg.ylabel,
#             zaxis_title=rv.name,
#         ),
#     )

#     fig = dict(data=data, layout=layout)

#     return pn.pane.Plotly(fig, name=PlotTypes.cone_plotly)


class VolumePlot:
    def volume_plotly(self, pl_in: PlotInput) -> Optional[pn.panel]:
        if PlotFilter(
            float_range=VarRange(3, 3),
            cat_range=VarRange(-1, 0),
            vector_len=VarRange(1, 1),
            result_vars=VarRange(1, 1),
        ).matches(pl_in.plt_cnt_cfg):
            sns_cfg = PltCfgBase()
            sns_cfg.y = pl_in.rv.name
            xr_cfg = plot_float_cnt_3(sns_cfg, pl_in.plt_cnt_cfg, pl_in.bench_cfg.debug)
            return plot_volume_plotly(pl_in.bench_cfg, pl_in.rv, xr_cfg)
        return None

    # def cone_plotly(self, pl_in: PlotInput) -> Optional[pn.panel]:
    #     if PlotFilter(
    #         float_range=VarRange(3, 3),
    #         cat_range=VarRange(-1, 0),
    #         vector_len=VarRange(3, 3),
    #         result_vars=VarRange(1, 1),
    #     ).matches(pl_in.plt_cnt_cfg):
    #         sns_cfg = PltCfgBase()
    #         sns_cfg.y = pl_in.rv.name
    #         xr_cfg = plot_float_cnt_3(sns_cfg, pl_in.plt_cnt_cfg, pl_in.bench_cfg.debug)
    #         return plot_cone_plotly(pl_in.bench_cfg, pl_in.rv, xr_cfg)
    #     return None
